"""Constraints for legacy implementations related to command parameters"""

from __future__ import annotations

from typing import (
    Any,
    Dict,
    TYPE_CHECKING,
    Type,
    TypeVar,
)

from .base import Constraint
from .basic import (
    EnsureBool,
    EnsureChoice,
    EnsureFloat,
    EnsureInt,
    EnsureStr,
    NoConstraint,
)
from .compound import (
    ConstraintWithPassthrough,
    EnsureIterableOf,
    EnsureMapping,
)
from .parameter import NoValue

if TYPE_CHECKING:  # pragma: no cover
    from datalad_next.commands import Parameter

EnsureParameterConstraint_T = TypeVar(
    'EnsureParameterConstraint_T',
    bound='EnsureParameterConstraint',
)


class EnsureParameterConstraint(EnsureMapping):
    """Ensures a mapping from a Python parameter name to a value constraint

    An optional "pass-though" value can be declare that is then exempt from
    validation and is returned as-is. This can be used to support, for example,
    special default values that only indicate the optional nature of a
    parameter. Declaring them as "pass-through" avoids a needless
    complexity-increase of a value constraint that would translate onto
    user-targeted error reporting.
    """
    # valid parameter name for Python and CLI
    # - must start with a lower-case letter
    # - must not contain symbols other than lower-case letters,
    #   digits, and underscore
    valid_param_name_regex = r'[a-z]{1}[a-z0-9_]*'

    def __init__(self,
                 constraint: Constraint,
                 passthrough: Any = NoValue):
        """
        Parameters
        ----------
        constraint:
          Any ``Constraint`` subclass instance that will be used to validate
          parameter values.
        passthrough:
          A value that will not be subjected to validation by the value
          constraint, but is returned as-is. This can be used to exempt
          default values from validation, e.g. when defaults are only
          placeholder values to indicate the optional nature of a parameter.
        """
        super().__init__(
            key=EnsureStr(
                match=EnsureParameterConstraint.valid_param_name_regex),
            value=ConstraintWithPassthrough(
                constraint,
                passthrough,
            ),
            # make it look like dict(...)
            delimiter='=',
        )

    @property
    def parameter_constraint(self):
        return self._value_constraint

    @property
    def passthrough_value(self):
        return self._value_constraint.passthrough

    def __call__(self, value) -> Dict:
        key, val = self._get_key_value(value)
        key = self._key_constraint(key)
        val = self._value_constraint(val) \
            if val != self.passthrough_value else val
        return {key: val}

    @classmethod
    def from_parameter(
            cls: Type[EnsureParameterConstraint_T],
            spec: Parameter,
            default: Any,
            item_constraint: Constraint | None = None,
            nargs: str | int | None = None) -> EnsureParameterConstraint_T:
        """
        Parameters
        ----------
        spec: Parameter
          Instance of a datalad-core Parameter. If not overwritten by values
          given to the other arguments of this method, item constraints,
          number of arguments and other argparse-specific information
          is taken from this object and processed to built a comprehensive
          constraint that handles all aspects of the specification in a
          homogeneous fashion via the Constraint interface.
        default: Any
          A parameter's default value. It is configured as a "pass-through"
          value that will not be subjected to validation.
        item_constraint:
          If given, it override any constraint declared in the Parameter
          instance given to `spec`
        nargs:
          If given, it override any nargs setting declared in the Parameter
          instance given to `spec`.
        """
        value_constraint = _get_comprehensive_constraint(
            spec,
            item_constraint,
            nargs,
        )
        return cls(value_constraint, passthrough=default)


# that mapping is NOT to be expanded!
# it is a legacy leftover. It's usage triggers a DeprecationWarning
_constraint_spec_map = {
    'float': EnsureFloat(),
    'int': EnsureInt(),
    'bool': EnsureBool(),
    'str': EnsureStr(),
}


def _get_comprehensive_constraint(
        param_spec: Parameter,
        # TODO remove `str` when literal constraint support is removed
        item_constraint_override: Constraint | str | None = None,
        nargs_override: str | int | None = None):
    action = param_spec.cmd_kwargs.get('action')
    # definitive per-item constraint, consider override
    # otherwise fall back on Parameter.constraints
    constraint = item_constraint_override or param_spec.constraints

    if not (constraint is None or hasattr(constraint, '__call__')):
        import warnings
        warnings.warn("Literal constraint labels are no longer supported.",
                      DeprecationWarning)
        try:
            return _constraint_spec_map[constraint]
        except KeyError:
            raise ValueError(
                f"unsupported constraint specification '{constraint}'")

    if not constraint:
        if action in ('store_true', 'store_false'):
            constraint = EnsureBool()
        elif param_spec.cmd_kwargs.get('choices'):
            constraint = EnsureChoice(*param_spec.cmd_kwargs.get('choices'))
        else:
            # always have one for simplicity
            constraint = NoConstraint()

    # we must additionally consider the following nargs spec for
    # a complete constraint specification
    # (int, '*', '+'), plus action=
    # - 'store_const' TODO
    # - 'store_true' and 'store_false' TODO
    # - 'append'
    # - 'append_const' TODO
    # - 'count' TODO
    # - 'extend' TODO

    # get the definitive argparse "nargs" value
    nargs = nargs_override or param_spec.cmd_kwargs.get('nargs', None)
    # try making a specific number explicit via dtype change
    try:
        nargs = int(nargs)
    except (ValueError, TypeError):
        pass

    # TODO reconsider using `list`, with no length-check it could
    # be a generator
    if isinstance(nargs, int):
        # we currently consider nargs=1 to be a request of a
        # single item, not a forced single-item list
        if nargs > 1:
            # sequence of a particular length
            constraint = EnsureIterableOf(
                list, constraint, min_len=nargs, max_len=nargs)
    elif nargs == '*':
        # datalad expects things often/always to also work for a single item
        constraint = EnsureIterableOf(list, constraint) | constraint
    elif nargs == '+':
        # sequence of at least 1 item, always a sequence,
        # but again datalad expects things often/always to also work for
        # a single item
        constraint = EnsureIterableOf(
            list, constraint, min_len=1) | constraint
    # handling of `default` and `const` would be here
    #elif nargs == '?'

    if action == 'append':
        # wrap into a(nother) sequence
        # (think: list of 2-tuples, etc.
        constraint = EnsureIterableOf(list, constraint)

    return constraint



"""Handler for operations, such as "download", on file:// URLs"""

# allow for |-type UnionType declarations
from __future__ import annotations

import logging
import random
import stat
import sys
import time
from io import IOBase
from math import floor
from pathlib import Path
from typing import (
    BinaryIO,
    Dict,
)
from urllib import (
    request,
    parse,
)

from datalad_next.consts import COPY_BUFSIZE

from .base import UrlOperations
from .exceptions import (
    UrlOperationsRemoteError,
    UrlOperationsResourceUnknown,
)

lgr = logging.getLogger('datalad.ext.next.file_url_operations')


__all__ = ['FileUrlOperations']


class FileUrlOperations(UrlOperations):
    """Handler for operations on `file://` URLs

    Access to local data via file-scheme URLs is supported with the
    same API and feature set as other URL-schemes (simultaneous
    content hashing and progress reporting.
    """
    def _file_url_to_path(self, url):
        assert url.startswith('file://')
        parsed = parse.urlparse(url)
        path = request.url2pathname(parsed.path)
        return Path(path)

    def stat(self,
             url: str,
             *,
             credential: str | None = None,
             timeout: float | None = None) -> Dict:
        """Gather information on a URL target, without downloading it

        See :meth:`datalad_next.url_operations.UrlOperations.stat`
        for parameter documentation and exception behavior.

        Raises
        ------
        UrlOperationsResourceUnknown
          For access targets found absent.
        """
        # filter out internals
        return {
            k: v for k, v in self._stat(url, credential).items()
            if not k.startswith('_')
        }

    def _stat(self, url: str, credential: str | None = None) -> Dict:
        # turn url into a native path
        from_path = self._file_url_to_path(url)
        # if anything went wrong with the conversion, or we lack
        # permissions: die here
        try:
            size = from_path.stat().st_size
        except FileNotFoundError as e:
            raise UrlOperationsResourceUnknown(url) from e
        return {
            'content-length': size,
            '_path': from_path,
        }

    def download(self,
                 from_url: str,
                 to_path: Path | None,
                 *,
                 # unused, but theoretically could be used to
                 # obtain escalated/different privileges on a system
                 # to gain file access
                 credential: str | None = None,
                 hash: list[str] | None = None,
                 timeout: float | None = None) -> Dict:
        """Copy a file:// URL target to a local path

        See :meth:`datalad_next.url_operations.UrlOperations.download`
        for parameter documentation and exception behavior.

        Raises
        ------
        UrlOperationsResourceUnknown
          For download targets found absent.
        """
        dst_fp = None
        try:
            props = self._stat(from_url, credential=credential)
            from_path = props['_path']
            expected_size = props['content-length']
            dst_fp = sys.stdout.buffer if to_path is None \
                else open(to_path, 'wb')

            with from_path.open('rb') as src_fp:
                props.update(self._copyfp(
                    src_fp,
                    dst_fp,
                    expected_size,
                    hash,
                    start_log=('Download %s to %s', from_url, to_path),
                    update_log=('Downloaded chunk',),
                    finish_log=('Finished download',),
                    progress_label='downloading',
                ))
                return props
        except PermissionError:
            # would be a local issue, pass-through
            raise
        except UrlOperationsResourceUnknown:
            # would come from stat(), pass_through
            raise
        except Exception as e:
            # wrap this into the datalad-standard, but keep the
            # original exception linked
            raise UrlOperationsRemoteError(from_url, message=str(e)) from e
        finally:
            if dst_fp and to_path is not None:
                dst_fp.close()

    def upload(self,
               from_path: Path | None,
               to_url: str,
               *,
               credential: str | None = None,
               hash: list[str] | None = None,
               timeout: float | None = None) -> Dict:
        """Copy a local file to a file:// URL target

        Any missing parent directories of the URL target are created as
        necessary.

        See :meth:`datalad_next.url_operations.UrlOperations.upload`
        for parameter documentation and exception behavior.

        Raises
        ------
        FileNotFoundError
          If the source file cannot be found.
        """
        # get the size, or die if inaccessible
        props = {}
        if from_path:
            expected_size = from_path.stat().st_size
            props['content-length'] = expected_size
        else:
            expected_size = None
        final_path = self._file_url_to_path(to_url)
        # added a `nosec` below because we don't need a cryptographically
        # secure random number here.
        time_stamp = time.time()
        to_path = final_path.with_suffix(   # nosec
            final_path.suffix
            + f'.transfer-{random.randint(1000000000, 9999999999)}'
            + f'_{str(time_stamp - floor(time_stamp))[2:]}'
        )
        # create parent dir(s) as necessary
        to_path.parent.mkdir(exist_ok=True, parents=True)
        src_fp = None
        try:
            src_fp = sys.stdin.buffer if from_path is None \
                else open(from_path, 'rb')
            with to_path.open('wb') as dst_fp:
                props.update(self._copyfp(
                    src_fp,
                    dst_fp,
                    expected_size,
                    hash,
                    start_log=('Upload %s to %s', from_path, to_url),
                    update_log=('Uploaded chunk',),
                    finish_log=('Finished upload',),
                    progress_label='uploading',
                ))
            to_path.replace(final_path)
            return props
        except FileNotFoundError as e:
            raise UrlOperationsResourceUnknown(from_path) from e
        except Exception as e:
            # wrap this into the datalad-standard, but keep the
            # original exception linked
            raise UrlOperationsRemoteError(to_url, message=str(e)) from e
        finally:
            if src_fp and from_path is not None:
                src_fp.close()

    def delete(self,
               url: str,
               *,
               credential: str | None = None,
               timeout: float | None = None) -> Dict:
        """Delete the target of a file:// URL

        The target can be a file or a directory. `delete` will attempt to
        delete write protected targets (by setting write permissions). If
        the target is a directory, the complete directory and all its
        content will be deleted. `delete` will not modify the permissions
        of the parent of the target. That means, it will not delete a target
        in a write protected directory, but it will empty target, if target is
        a directory.

        See :meth:`datalad_next.url_operations.UrlOperations.delete`
        for parameter documentation and exception behavior.

        Raises
        ------
        UrlOperationsResourceUnknown
          For deletion targets found absent.
        """
        path = self._file_url_to_path(url)
        try:
            if path.is_dir():
                self._delete_dir(path)
            else:
                path.chmod(stat.S_IWUSR)
                path.unlink()
        except FileNotFoundError as e:
            raise UrlOperationsResourceUnknown(url) from e
        except Exception as e:
            # wrap this into the datalad-standard, but keep the
            # original exception linked
            raise UrlOperationsRemoteError(url, message=str(e)) from e
        return {}

    def _delete_dir(self, path: Path):
        path.chmod(stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)
        for sub_path in path.iterdir():
            if sub_path.is_dir():
                self._delete_dir(sub_path)
            else:
                sub_path.chmod(stat.S_IWUSR)
                sub_path.unlink()
        path.rmdir()

    def _copyfp(self,
                src_fp: IOBase | BinaryIO,
                dst_fp: IOBase | BinaryIO,
                expected_size: int | None,
                hash: list[str] | None,
                start_log: tuple,
                update_log: tuple,
                finish_log: tuple,
                progress_label: str,
    ) -> dict:
        # this is pretty much shutil.copyfileobj() with the necessary
        # wrapping to perform hashing and progress reporting
        hasher = self._get_hasher(hash)
        progress_id = self._get_progress_id(str(id(src_fp)), str(id(dst_fp)))

        # Localize variable access to minimize overhead
        src_fp_read = src_fp.read
        dst_fp_write = dst_fp.write

        props = {}
        self._progress_report_start(
            progress_id, start_log, progress_label, expected_size)
        copy_size = 0
        try:
            while True:
                chunk = src_fp_read(COPY_BUFSIZE)
                if not chunk:
                    break
                dst_fp_write(chunk)
                chunk_size = len(chunk)
                self._progress_report_update(
                    progress_id, update_log, chunk_size)
                # compute hash simultaneously
                hasher.update(chunk)
                copy_size += chunk_size
            props.update(hasher.get_hexdigest())
            # return how much was copied. we could compare with
            # `expected_size` and error on mismatch, but not all
            # sources can provide that (e.g. stdin)
            props['content-length'] = copy_size
            return props
        finally:
            self._progress_report_stop(progress_id, finish_log)



import pytest

from ..base import (
    Constraint,
    AllOf,
    AnyOf,
)
from ..basic import (
    EnsureDType,
    EnsureInt,
    EnsureFloat,
    EnsureBool,
    EnsureNone,
    EnsureRange,
    EnsureStr,
)


def test_base():
    # there is no "standard" implementation
    with pytest.raises(NotImplementedError):
        Constraint()('whatever')
    # no standard docs
    with pytest.raises(NotImplementedError):
        Constraint().short_description()
    with pytest.raises(NotImplementedError):
        Constraint().long_description()
    # dataset context switching is by default a no-op
    generic = Constraint()
    assert id(generic) == id(generic.for_dataset('some'))


def test_allof():
    # this should always work
    c = AllOf(EnsureFloat())
    assert c(7.0) == 7.0
    c = AllOf(EnsureFloat(), EnsureRange(min=4.0))
    assert repr(c) == 'AllOf(EnsureFloat(), EnsureRange())'
    assert c(7.0) == 7.0
    # __and__ form
    c = EnsureFloat() & EnsureRange(min=4.0)
    assert c.short_description() == '(float and not less than 4.0)'
    assert 'and not less than 4.0' in c.long_description()
    assert c(7.0) == 7.0
    with pytest.raises(ValueError):
        c(3.9)
    c = AllOf(EnsureFloat(), EnsureRange(min=4), EnsureRange(max=9))
    assert c(7.0) == 7.0
    with pytest.raises(ValueError):
        c(3.9)
    with pytest.raises(ValueError):
        c(9.01)
    # __and__ form
    c = EnsureFloat() & EnsureRange(min=4) & EnsureRange(max=9)
    assert c(7.0) == 7.0
    with pytest.raises(ValueError):
        c(3.99)
    with pytest.raises(ValueError):
        c(9.01)
    # and reordering should not have any effect
    c = AllOf(EnsureRange(max=4), EnsureRange(min=9), EnsureFloat())
    with pytest.raises(ValueError):
        c(3.99)
    with pytest.raises(ValueError):
        c(9.01)
    # smoke test concat AND constraints
    c1 = AllOf(EnsureRange(max=10), EnsureRange(min=5))
    c2 = AllOf(EnsureRange(max=6), EnsureRange(min=2))
    c = c1 & c2
    # make sure that neither c1, nor c2 is modified
    assert len(c1.constraints) == 2
    assert len(c2.constraints) == 2
    assert len(c.constraints) == 4
    assert c(6) == 6
    with pytest.raises(ValueError):
        c(4)


def test_anyof():
    # this should always work
    c = AnyOf(EnsureFloat())
    # passes the docs through
    assert c.short_description() == EnsureFloat().short_description()
    assert c(7.0) == 7.0
    c = AnyOf(EnsureFloat(), EnsureNone())
    assert repr(c) == 'AnyOf(EnsureFloat(), EnsureNone())'
    # wraps docs in parenthesis to help appreciate the scope of the
    # OR'ing
    assert c.short_description().startswith(
        f'({EnsureFloat().short_description()}')
    assert c.short_description(), '(float or None)'
    assert c(7.0) == 7.0
    assert c(None) is None
    # OR with an alternative just extends
    c = c | EnsureInt()
    assert c.short_description(), '(float or None or int)'
    # OR with an alternative combo also extends
    c = c | AnyOf(EnsureBool(), EnsureInt())
    # yes, no de-duplication
    assert c.short_description(), '(float or None or int or bool or int)'
    # spot check long_description, must have some number
    assert len(c.long_description().split(' or ')) == 5
    # __or__ form
    c = EnsureFloat() | EnsureNone()
    assert c(7.0) == 7.0
    assert c(None) is None

    # this should always fail
    c = AllOf(EnsureRange(min=0, max=4), EnsureRange(min=9, max=11))
    with pytest.raises(ValueError):
        c(7.0)
    c = EnsureRange(min=0, max=4) | EnsureRange(min=9, max=11)
    assert c(3.0) == 3.0
    assert c(9.0) == 9.0
    with pytest.raises(ValueError):
        c(7.0)
    with pytest.raises(ValueError):
        c(-1.0)

    # verify no inplace modification
    c1 = EnsureInt() | EnsureStr()
    c2 = c1 | EnsureDType(c1)
    # OR'ing does not "append" the new alternative to c1.
    assert len(c1.constraints) == 2
    # at the same time, c2 does not contain an AnyOf
    # as an internal constraint, because this would be needless
    # complexity re the semantics of OR
    assert len(c2.constraints) == 3


def test_both():
    # this should always work
    c = AnyOf(
        AllOf(
            EnsureFloat(),
            EnsureRange(min=7.0, max=44.0)),
        EnsureNone(),
    )
    assert c(7.0) == 7.0
    assert c(None) is None
    # this should always fail
    with pytest.raises(ValueError):
        c(77.0)



"""Report on the status of the worktree

The main functionality is provided by the :func:`iter_gitstatus` function.
"""
from __future__ import annotations

import logging
from pathlib import Path
from typing import (
    Iterator,
    Generator,
)

from datalad_next.consts import PRE_INIT_COMMIT_SHA
from datalad_next.runners import (
    call_git_lines,
)
from datalad_next.repo_utils import (
    get_worktree_head,
)

from .gitdiff import (
    GitDiffItem,
    GitDiffStatus,
    GitContainerModificationType,
    iter_gitdiff,
)
from .gitworktree import (
    GitTreeItemType,
    iter_gitworktree,
    iter_submodules,
)

lgr = logging.getLogger('datalad.ext.next.iter_collections.gitstatus')

# map untracked modes given to iter_gitstatus() to those that need to be
# passed to iter_gitworktree() for getting untracked content only
_untracked_mode_map = {
    'all': 'only',
    'whole-dir': 'only-whole-dir',
    'no-empty-dir': 'only-no-empty-dir',
}


def iter_gitstatus(
    path: Path,
    *,
    untracked: str | None = 'all',
    recursive: str = 'repository',
    eval_submodule_state: str = "full",
) -> Generator[GitDiffItem, None, None]:
    """
    Recursion mode 'no'

    This mode limits the reporting to immediate directory items of a given
    path. This mode is not necessarily faster than a 'repository' recursion.
    Its primary purpose is the ability to deliver a collapsed report in that
    subdirectories are treated similar to submodules -- as containers that
    maybe have modified or untracked content.

    Parameters
    ----------
    path: Path
      Path of a directory in a Git repository to report on. This directory
      need not be the root directory of the repository, but must be part of
      the repository. If the directory is not the root directory of a
      non-bare repository, the iterator is constrained to items underneath
      that directory.
    untracked: {'all', 'whole-dir', 'no-empty-dir'} or None, optional
      If not ``None``, also reports on untracked work tree content.
      ``all`` reports on any untracked file; ``whole-dir`` yields a single
      report for a directory that is entirely untracked, and not individual
      untracked files in it; ``no-empty-dir`` skips any reports on
      untracked empty directories. Also see ``eval_submodule_state`` for
      how this parameter is applied in submodule recursion.
    recursive: {'no', 'repository', 'submodules', 'monolithic'}, optional
      Behavior for recursion into subtrees. By default (``repository``),
      all trees within the repository underneath ``path``) are reported,
      but no tree within submodules. With ``submodules``, recursion includes
      any submodule that is present. If ``no``, only direct children
      are reported on.
    eval_submodule_state: {"no", "commit", "full"}, optional
      If 'full' (default), the state of a submodule is evaluated by
      considering all modifications, with the treatment of untracked files
      determined by `untracked`. If 'commit', the modification check is
      restricted to comparing the submodule's "HEAD" commit to the one
      recorded in the superdataset. If 'no', the state of the subdataset is
      not evaluated. When a git-annex repository in adjusted mode is detected,
      the reference commit that the worktree is being compared to is the basis
      of the adjusted branch (i.e., the corresponding branch).

    Yields
    ------
    :class:`GitDiffItem`
      The ``name`` and ``prev_name`` attributes of an item are a ``str`` with
      the corresponding (relative) path, as reported by Git
      (in POSIX conventions).

    .. note::

       The implementation requires `git rev-parse --path-format=relative`
       that was introduced with Git v2.31.
    """
    path = Path(path)

    head, corresponding_head = get_worktree_head(path)
    if head is None:
        # no commit at all -> compare to an empty repo.
        head = PRE_INIT_COMMIT_SHA

    # TODO it would make sense to always (or optionally) compare against any
    # existing corresponding_head. This would make the status communicate
    # anything that has not made it into the corresponding branch yet

    common_args = dict(
        head=head,
        path=path,
        untracked=untracked,
        eval_submodule_state=eval_submodule_state,
    )

    if recursive == 'no':
        yield from _yield_dir_items(**common_args)
        return
    elif recursive == 'repository':
        yield from _yield_repo_items(**common_args)
    # TODO what we really want is a status that is not against a per-repository
    # HEAD, but against the commit that is recorded in the parent repository
    # TODO we need a name for that
    elif recursive in ('submodules', 'monolithic'):
        yield from _yield_hierarchy_items(
            recursion_mode=recursive,
            **common_args,
        )
    else:
        raise ValueError(f'unknown recursion type {recursive!r}')


#
# status generators for each mode
#

def _yield_dir_items(
    *,
    head: str | None,
    path: Path,
    untracked: str | None,
    eval_submodule_state: str,
) -> Iterator[GitDiffItem]:
    # potential container items in a directory that need content
    # investigation
    container_types = (
        GitTreeItemType.directory,
        GitTreeItemType.submodule,
    )
    if untracked is None:
        # no need to look at anything other than the diff report
        dir_items = {}
    else:
        # there is no recursion, avoid wasting cycles on listing individual
        # files in subdirectories
        untracked = 'whole-dir' if untracked == 'all' else untracked
        # gather all directory items upfront, we subtract the ones reported
        # modified later and lastly yield all untracked content from them
        dir_items = {
            item.name.as_posix(): item
            for item in iter_gitworktree(
                path,
                untracked=untracked,
                recursive='no',
            )
        }
    # diff constrained to direct children
    for item in iter_gitdiff(
        path,
        from_treeish=head,
        # to the worktree
        to_treeish=None,
        recursive='no',
        # TODO trim scope like in repo_items
        eval_submodule_state=eval_submodule_state,
    ):
        if item.status != GitDiffStatus.deletion \
                and item.gittype in container_types:
            if item.gittype == GitTreeItemType.submodule:
                # issue standard submodule container report
                _eval_submodule(path, item, eval_submodule_state)
            else:
                dir_path = path / item.path
                # this is on a directory. if it appears here, it has
                # modified content
                if dir_path.exists():
                    item.add_modification_type(
                        GitContainerModificationType.modified_content)
                    if untracked is not None \
                            and _path_has_untracked(dir_path):
                        item.add_modification_type(
                            GitContainerModificationType.untracked_content)
                else:
                    # this directory is gone entirely
                    item.status = GitDiffStatus.deletion
                    item.modification_types = None
            # we dealt with this item completely
            dir_items.pop(item.name, None)
        if item.status:
            yield item

    if untracked is None:
        return

    # yield anything untracked, and inspect remaining containers
    for dir_item in dir_items.values():
        if dir_item.gitsha is None and dir_item.gittype is None:
            # this is untracked
            yield GitDiffItem(
                # for homgeneity for report a str-path no matter what
                name=str(dir_item.name),
                status=GitDiffStatus.other,
            )
        elif dir_item.gittype in container_types:
            # none of these containers has any modification other than
            # possibly untracked content
            item = GitDiffItem(
                # for homgeneity for report a str-path no matter what
                name=str(dir_item.name),
                # this submodule has not been detected as modified
                # per-commit, assign reported gitsha to pre and post
                # state
                gitsha=dir_item.gitsha,
                prev_gitsha=dir_item.gitsha,
                gittype=dir_item.gittype,
                # TODO others?
            )
            if item.gittype == GitTreeItemType.submodule:
                # issue standard submodule container report
                _eval_submodule(path, item, eval_submodule_state)
            else:
                # this is on a directory. if it appears here, it has
                # no modified content
                testpath = path / dir_item.path
                if _path_has_untracked(path / dir_item.path):
                    item.status = GitDiffStatus.modification
                    item.add_modification_type(
                        GitContainerModificationType.untracked_content)
            if item.status:
                yield item


def _yield_repo_items(
    *,
    head: str | None,
    path: Path,
    untracked: str | None,
    eval_submodule_state: str,
) -> Generator[GitDiffItem, None, None]:
    """Report status items for a single/whole repsoitory"""
    present_submodules = {
        # stringify name for speedy comparison
        # TODO double-check that comparisons are primarily with
        # GitDiffItem.name which is str
        str(item.name): item for item in iter_submodules(path)
    }
    # start with a repository-contrained diff against the worktree
    for item in iter_gitdiff(
        path,
        from_treeish=head,
        # to the worktree
        to_treeish=None,
        recursive='repository',
        # we should be able to go cheaper with the submodule evaluation here.
        # We need to redo some check for adjusted mode, and other cases anyways
        eval_submodule_state='commit'
        if eval_submodule_state == 'full' else eval_submodule_state,
    ):
        # immediately investigate any submodules that are already
        # reported modified by Git
        if item.gittype == GitTreeItemType.submodule:
            _eval_submodule(path, item, eval_submodule_state)
            # we dealt with this submodule
            present_submodules.pop(item.name, None)
        if item.status:
            yield item

    # we are not generating a recursive report for submodules, hence
    # we need to look at ALL submodules for untracked content
    # `or {}` for the case where we got no submodules, which happens
    # with `eval_submodule_state == 'no'`
    for subm_name, subm_item in (present_submodules or {}).items():
        # none of these submodules has any modification other than
        # possibly untracked content
        item = GitDiffItem(
            # for homgeneity for report a str-path no matter what
            name=str(subm_item.name),
            # this submodule has not been detected as modified
            # per-commit, assign reported gitsha to pre and post
            # state
            gitsha=subm_item.gitsha,
            prev_gitsha=subm_item.gitsha,
            gittype=subm_item.gittype,
            # TODO others?
        )
        # TODO possibly trim eval_submodule_state
        _eval_submodule(path, item, eval_submodule_state)
        if item.status:
            yield item

    if untracked is None:
        return

    # lastly untracked files of this repo
    for untracked_item in iter_gitworktree(
        path=path,
        untracked=_untracked_mode_map[untracked],
        link_target=False,
        fp=False,
        recursive='repository',
    ):
        yield GitDiffItem(
            name=untracked_item.name.as_posix(),
            status=GitDiffStatus.other,
            gittype=untracked_item.gittype,
        )


def _yield_hierarchy_items(
    *,
    head: str | None,
    path: Path,
    untracked: str | None,
    recursion_mode: str,
    eval_submodule_state: str,
) -> Generator[GitDiffItem, None, None]:
    for item in _yield_repo_items(
        head=head,
        path=path,
        untracked=untracked,
        # TODO do we need to adjust the eval mode here for the diff recmodes?
        eval_submodule_state=eval_submodule_state,
    ):
        # there is nothing else to do for any non-submodule item
        if item.gittype != GitTreeItemType.submodule:
            yield item
            continue

        # we get to see any submodule item passing through here, and can simply
        # call this function again for a subpath

        # submodule recursion
        # the .path of a GitTreeItem is always POSIX
        sm_path = path / item.path
        if recursion_mode == 'submodules':
            # in this mode, we run the submodule status against it own
            # worktree head
            sm_head, _ = get_worktree_head(sm_path)
            # because this need not cover all possible changes with respect
            # to the parent repository, we yield an item on the submodule
            # itself
            yield item
        elif recursion_mode == 'monolithic':
            # in this mode we determine the change of the submodule with
            # respect to the recorded state in the parent. This is either
            # the current gitsha, or (if git detected a committed
            # modification) the previous sha. This way, any further report
            # on changes a comprehensive from the point of view of the parent
            # repository, hence no submodule item is emitted
            sm_head = item.gitsha or item.prev_gitsha

            if item.modification_types is None \
                    or GitContainerModificationType.new_commits \
                    in item.modification_types:
                # this is a submodule that is either entriely new (added),
                # or it has new commits compared to
                # its state in the parent dataset. We need to yield this
                # item, even if nothing else is modified, because otherwise
                # this (unsafed) changed would go unnoticed
                # https://github.com/datalad/datalad-next/issues/645
                yield item

        for i in _yield_hierarchy_items(
            head=sm_head,
            path=sm_path,
            untracked=untracked,
            # TODO here we could implement handling for a recursion-depth limit
            recursion_mode=recursion_mode,
            eval_submodule_state=eval_submodule_state,
        ):
            i.name = f'{item.name}/{i.name}'
            yield i


#
# Helpers
#


def _path_has_untracked(path: Path) -> bool:
    """Recursively check for any untracked content (except empty dirs)"""
    if not path.exists():
        # cannot possibly have untracked
        return False
    for ut in iter_gitworktree(
        path=path,
        untracked='only-no-empty-dir',
        link_target=False,
        fp=False,
        recursive='submodules',
    ):
        # fast exit on the first detection
        return True
    # only after we saw everything we can say there is nothing
    return False


def _get_submod_worktree_head(path: Path) -> tuple[bool, str | None, bool]:
    """Returns (submodule exists, SHA | None, adjusted)"""
    try:
        HEAD, corresponding_head = get_worktree_head(path)
    except ValueError:
        return False, None, False

    adjusted = corresponding_head is not None
    if adjusted:
        # this is a git-annex adjusted branch. do the comparison against
        # its basis. it is not meaningful to track the managed branch in
        # a superdataset
        HEAD = corresponding_head
    res = call_git_lines(
        ['rev-parse', '--path-format=relative', '--show-toplevel', HEAD],
        cwd=path,
    )
    assert len(res) == 2
    if res[0].startswith('..'):
        # this is not a report on a submodule at this location
        return False, None, adjusted
    else:
        return True, res[1], adjusted


def _eval_submodule(
    basepath: Path,
    item: GitDiffItem,
    eval_mode: str,
) -> None:
    """In-place amend GitDiffItem submodule item

    It does nothing with ``eval_mode='no'``.
    """
    if eval_mode == 'no':
        return

    item_path = basepath / item.path

    # this is the cheapest test for the theoretical chance that a submodule
    # is present at `item_path`. This is beneficial even when we would only
    # run a single call to `git rev-parse`
    # https://github.com/datalad/datalad-next/issues/606
    if not (item_path / '.git').exists():
        return

    # get head commit, and whether a submodule is actually present,
    # and/or in adjusted mode
    subds_present, head_commit, adjusted = _get_submod_worktree_head(item_path)
    if not subds_present:
        return

    if adjusted:
        _eval_submodule_adjusted(item_path, item, head_commit, eval_mode)
    else:
        _eval_submodule_normal(item_path, item, head_commit, eval_mode)


def _eval_submodule_normal(
    item_path: Path,
    item: GitDiffItem,
    head_commit: str | None,
    eval_mode: str,
) -> None:
    if eval_mode == 'full' and item.status is None or (
        item.modification_types
        and GitContainerModificationType.new_commits in item.modification_types
    ):
        # if new commits have been detected, the diff-implementation is
        # not able to report "modified content" at the same time, if it
        # exists. This requires a dedicated inspection, which conincidentally
        # is identical to the analysis of an adjusted mode submodule.
        return _eval_submodule_adjusted(
            item_path, item, head_commit, eval_mode)

    if item.gitsha != head_commit:
        item.status = GitDiffStatus.modification
        item.add_modification_type(GitContainerModificationType.new_commits)

    if eval_mode == 'commit':
        return

    # check for untracked content (recursively)
    if _path_has_untracked(item_path):
        item.status = GitDiffStatus.modification
        item.add_modification_type(
            GitContainerModificationType.untracked_content)


def _eval_submodule_adjusted(
    item_path: Path,
    item: GitDiffItem,
    head_commit: str | None,
    eval_mode: str,
) -> None:
    # we cannot rely on the diff-report for a submodule in adjusted mode.
    # git would make the comparison to the adjusted branch HEAD alone.
    # this would almost always be invalid, because it is not meaningful to
    # track a commit in an adjusted branch (it goes away).
    #
    # instead, we need to:
    # - check for a change in the corresponding HEAD to the recorded commit
    #   in the parent repository, consider any change "new commits"
    # - check for a diff of the worktree to corresponding HEAD, consider
    #   any such diff a "modified content"
    # - and lastly check for untracked content

    # start with "no modification"
    item.status = None
    item.modification_types = None

    if item.prev_gitsha != head_commit:
        item.status = GitDiffStatus.modification
        item.add_modification_type(GitContainerModificationType.new_commits)

    if eval_mode == 'commit':
        return

    if any(
        i.status is not None
        for i in iter_gitdiff(
            item_path,
            from_treeish=head_commit,
            # worktree
            to_treeish=None,
            recursive='repository',
            find_renames=None,
            find_copies=None,
            eval_submodule_state='commit',
        )
    ):
        item.status = GitDiffStatus.modification
        item.add_modification_type(
            GitContainerModificationType.modified_content)

    # check for untracked content (recursively)
    if _path_has_untracked(item_path):
        item.status = GitDiffStatus.modification
        item.add_modification_type(
            GitContainerModificationType.untracked_content)



"""Enable SSH-based remote command execution on Windows

This change introduces a replacement for core's
``datalad/support/sshconnector.py:BaseSSHConnection._exec_ssh()``
with a dedicated handling of ``stdin`` for Windows.

The OpenSSH client in Windows modifies its ``stdin``-descriptor in such a way, that it becomes unusable for the python process, if the ``stdin``-descriptor is shared between the ``python``-process and the ``ssh``-process. As a result, all read-operations that the ``python``-process performs on ``stdin`` will block and leave the python-process "hanging".

This change passes an explicit, empty, byte-string as ``stdin`` to the
SSH client call, in order to avoid any interaction of SSH with the
parent process's ``stdin`` descriptor.
"""

import logging

from datalad.support.sshconnector import (
    StdOutErrCapture,
    NoCapture,
)

from datalad_next.patches import apply_patch
from datalad_next.utils import on_windows

# use same logger as -core
lgr = logging.getLogger('datalad.support.sshconnector')


# This method interface/original implementation is taken from
# datalad-core@58b8e06317fe1a03290aed80526bff1e2d5b7797
# datalad/support/sshconnector.py:BaseSSHConnection
def _exec_ssh(self, ssh_cmd, cmd, options=None, stdin=None, log_output=True):
    cmd = self._adjust_cmd_for_bundle_execution(cmd)

    for opt in options or []:
        ssh_cmd.extend(["-o", opt])

    # THIS IS THE PATCH
    if on_windows and stdin is None:
        # SSH on windows requires a special stdin handling. If we'd let
        # stdin=None do its normal thing, the Python process would hang,
        # because it looses touch with its own file descriptor.
        # See https://github.com/datalad/datalad-ria/issues/68
        stdin = b''

    # build SSH call, feed remote command as a single last argument
    # whatever it contains will go to the remote machine for execution
    # we cannot perform any sort of escaping, because it will limit
    # what we can do on the remote, e.g. concatenate commands with '&&'
    ssh_cmd += [self.sshri.as_str()] + [cmd]

    lgr.debug("%s is used to run %s", self, ssh_cmd)

    # TODO: pass expect parameters from above?
    # Hard to explain to toplevel users ... So for now, just set True
    out = self.runner.run(
        ssh_cmd,
        protocol=StdOutErrCapture if log_output else NoCapture,
        stdin=stdin)
    return out['stdout'], out['stderr']


apply_patch(
    'datalad.support.sshconnector', 'BaseSSHConnection', '_exec_ssh',
    _exec_ssh,
)



from pathlib import Path
from unittest.mock import (
    call,
    patch,
)
from urllib.parse import quote as urlquote

from datalad_next.tests import (
    assert_in,
    assert_in_results,
    assert_raises,
    assert_result_count,
    assert_status,
    create_tree,
    eq_,
    ok_,
    run_main,
)
import pytest

from datalad.api import (
    clone,
    create_sibling_webdav,
)
from datalad_next.utils import chpwd


def test_common_workflow_implicit_cred(
        credman, existing_dataset, tmp_path, webdav_credential, webdav_server,
        no_result_rendering):
    check_common_workflow(
        False, 'annex',
        credman, existing_dataset, tmp_path, webdav_credential, webdav_server)


def test_common_workflow_explicit_cred(
        credman, existing_dataset, tmp_path, webdav_credential, webdav_server,
        no_result_rendering):
    check_common_workflow(
        True, 'annex',
        credman, existing_dataset, tmp_path, webdav_credential, webdav_server)


def test_common_workflow_export(
        credman, existing_dataset, tmp_path, webdav_credential, webdav_server,
        no_result_rendering):
    check_common_workflow(
        False, 'filetree',
        credman, existing_dataset, tmp_path, webdav_credential, webdav_server)


def check_common_workflow(
        declare_credential, mode,
        credman, ds, clonepath, webdav_credential, webdav_server):
    credman.set(**webdav_credential)
    # need to amend the test credential, can only do after we know the URL
    ds.credentials(
        'set',
        name=webdav_credential['name'],
        # the test webdav webserver uses a realm label '/'
        spec=dict(realm=webdav_server.url + '/'),
    )

    # we use a nasty target directory that has the potential to ruin the
    # git-remote URL handling
    targetdir_name = 'tar&get=mike'
    targetdir = Path(webdav_server.path) / targetdir_name
    url = f'{webdav_server.url}/{targetdir_name}'

    with chpwd(ds.path):
        res = create_sibling_webdav(
            url,
            credential=webdav_credential['name']
            if declare_credential else None,
            mode=mode,
        )
        # Ensure that remote name constraint check works
        # second time should raise because the sibling exists already
        with pytest.raises(ValueError) as e:
            create_sibling_webdav(
                url,
                credential=webdav_credential['name']
                if declare_credential else None,
                mode=mode,
                name='127.0.0.1',
            )
        with pytest.raises(ValueError) as e:
            create_sibling_webdav(
                url,
                credential=webdav_credential['name']
                if declare_credential else None,
                mode=mode,
                name='other',
                storage_name='127.0.0.1-storage',
            )

    assert_in_results(
        res,
        action='create_sibling_webdav.storage',
        status='ok',
        type='sibling',
        path=ds.path,
        url=url,
        name='127.0.0.1-storage',
        # TODO: name=???
        #  Parse url for host or depend on serve_path always
        #  delivering 127.0.0.1? (Think IPv6 or a literal `localhost` or
        #  anything like that) Same applies to hardcoded `dlaurl`.
    )
    # where it should be accessible
    # needs to be quoted
    dlaurl = (
        'datalad-annex::?type=webdav&encryption=none&exporttree={exp}&'
        'url=http%3A//127.0.0.1%3A43612/tar%26get%3Dmike').format(
        exp='yes' if 'filetree' in mode else 'no',
    )
    if declare_credential:
        dlaurl += f'&dlacredential={urlquote(webdav_credential["name"])}'

    assert_in_results(
        res,
        action='create_sibling_webdav',
        status='ok',
        path=ds.path,
        name='127.0.0.1',
        url=dlaurl,
        type='sibling',
    )
    ok_(targetdir.exists())
    # add some annex payload
    (ds.pathobj / 'testfile.dat').write_text('dummy')
    ds.save()
    res = ds.push(to='127.0.0.1')
    assert_in_results(
        res,
        action='copy',
        path=str(ds.pathobj / 'testfile.dat'),
        status='ok',
    )
    assert_in_results(res, action='publish', status='ok')

    cloneurl = dlaurl
    if not declare_credential and 'filetree' in mode:
        # we can use a simplified URL
        cloneurl = 'webdav://{url}'.format(
            # strip http://
            url=url[7:],
        )
    dsclone = clone(cloneurl, clonepath)
    # we get the same thing
    eq_(ds.repo.get_hexsha(ds.repo.get_corresponding_branch()),
        dsclone.repo.get_hexsha(dsclone.repo.get_corresponding_branch()))

    # check that it auto-deploys webdav credentials
    # at some point, clone should be able to do this internally
    # https://github.com/datalad/datalad/issues/6634
    dsclone.siblings('enable', name='127.0.0.1-storage')
    # verify that we can get testfile.dat
    # just get the whole damn thing
    assert_status('ok', dsclone.get('.'))
    # verify testfile content
    eq_('dummy', (dsclone.pathobj / 'testfile.dat').read_text())
    # ensure that recursive operations succeed
    # create a subdataset
    subds = ds.create('mysubds')
    targetdir_name = 'recursiontest'
    subtargetdir = Path(webdav_server.path) / targetdir_name / 'mysubds'
    url = f'{webdav_server.url}/{targetdir_name}'

    with chpwd(ds.path):
        res = create_sibling_webdav(
            url,
            credential=webdav_credential['name']
            if declare_credential else None,
            name='recursive-sibling',
            mode=mode,
            recursive=True,
        )
        assert len(res) == 4  # 2 for create-sibling-webdav, 2 for storage
    assert_in_results(
        res,
        action='create_sibling_webdav.storage',
        status='ok',
        type='sibling',
        path=subds.path,
        name='recursive-sibling-storage',
    )
    ok_(subtargetdir.exists())


def test_bad_url_catching(existing_dataset):
    # Ensure that bad URLs are detected and handled
    ds = existing_dataset
    check_pairs = [
        (
            "http://localhost:33322/abc?a",
            "URL has forbidden 'query' component"
        ),
        (
            "https://netloc/has-a-fragment#sdsd",
            "URL has forbidden 'fragment' component"
        ),
        (
            "https:///has-no-net-location",
            "URL is missing 'netloc' component"
        ),
        (
            "xxx://localhost:33322/abc",
            "url={url!r}\n  does not match expression '^(http|https)://'"
        ),
    ]

    for bad_url, expected_message in check_pairs:
        with pytest.raises(ValueError) as e:
            create_sibling_webdav(dataset=ds, url=bad_url)
        assert expected_message.format(url=bad_url) in str(e.value)


def test_http_warning(existing_dataset):
    # Check that usage of http: triggers a warning.
    ds = existing_dataset
    url = "http://localhost:33322/abc"

    with patch("datalad_next.commands.create_sibling_webdav._create_sibling_webdav") as csw_mock, \
         patch("datalad_next.commands.create_sibling_webdav.lgr") as lgr_mock:

        csw_mock.return_value = iter([])

        with pytest.raises(Exception):
            create_sibling_webdav(dataset=ds, url=url)

        eq_(lgr_mock.warning.call_count, 1)
        assert_in(
            call(
                f"Using 'http:' ({url!r}) means that WebDAV credentials are "
                f"sent unencrypted over network links. Consider using "
                f"'https:'."),
            lgr_mock.warning.mock_calls)


def test_constraints_checking(existing_dataset):
    # Ensure that constraints are checked internally
    ds = existing_dataset
    url = "http://localhost:22334/abc"
    for key in ("existing", "mode"):
        with pytest.raises(ValueError) as e:
            create_sibling_webdav(
                dataset=ds, url=url, **{key: "illegal-value"})
        assert "is not one of" in str(e.value)


def test_name_clash_detection(existing_dataset):
    # Ensure that constraints are checked internally
    ds = existing_dataset
    url = "http://localhost:22334/abc"
    for mode in ("annex", 'filetree', 'annex-only', 'filetree-only'):
        with pytest.raises(ValueError) as e:
            create_sibling_webdav(
                dataset=ds, url=url, name="abc", storage_name="abc", mode=mode)
        assert "sibling names must not be equal" in str(e.value)


def test_unused_storage_name_warning(existing_dataset):
    # Ensure that constraints are checked internally
    ds = existing_dataset
    url = "https://localhost:22334/abc"

    with patch("datalad_next.commands.create_sibling_webdav._create_sibling_webdav") as csw_mock, \
         patch("datalad_next.commands.create_sibling_webdav.lgr") as lgr_mock:

        csw_mock.return_value = iter([])

        mode_values = ("git-only", "annex-only", "filetree-only")
        for mode in mode_values:
            # We set up the mocks to generate the following exception. This allows
            # us to limit the test to the logic in 'create_sibling_wabdav'.
            assert_raises(
                Exception,
                create_sibling_webdav,
                dataset=ds,
                url=url,
                name="abc",
                storage_name="abc",
                mode=mode)
        eq_(lgr_mock.warning.call_count, len(mode_values))


def test_existing_switch(existing_dataset, credman, webdav_credential,
                         webdav_server, no_result_rendering):
    credman.set(**webdav_credential)
    check_existing_switch(existing_dataset, webdav_credential, webdav_server)


def check_existing_switch(ds, webdav_credential, webdav_server):
    create_tree(
        ds.path,
        {'sub': {'f0': '0'},
         'sub2': {'subsub': {'f1': '1'},
                  'f2': '2'},
         'f3': '3'}
    )
    # use a tricky name: '3f7' will be the hashdir of the XDLRA
    # key containing the superdataset's datalad-annex archive after a push
    sub = ds.create('3f7', force=True)
    sub2 = ds.create('sub2', force=True)
    subsub = sub2.create('subsub', force=True)
    ds.save(recursive=True)

    url = webdav_server.url
    # need to amend the test credential, can only do after we know the URL
    ds.credentials(
        'set',
        name=webdav_credential['name'],
        # the test webdav webserver uses a realm label '/'
        spec=dict(realm=url + '/'),
    )

    subsub.create_sibling_webdav(f'{url}/sub2/subsub', mode='annex')
    sub2.create_sibling_webdav(f'{url}/sub2', mode='annex-only')
    sub.create_sibling_webdav(f'{url}/3f7', mode='git-only')

    res = ds.create_sibling_webdav(url, mode='annex',
                                   existing='skip',
                                   recursive=True)
    dlaurl='datalad-annex::?type=webdav&encryption=none&exporttree=no&' \
           'url=http%3A//127.0.0.1%3A43612/'

    # results per dataset:
    # super:
    assert_in_results(
        res,
        action='create_sibling_webdav.storage',
        status='ok',
        type='sibling',
        name='127.0.0.1-storage',
        path=ds.path,
        url=url,
    )
    assert_in_results(
        res,
        action='create_sibling_webdav',
        status='ok',
        type='sibling',
        path=ds.path,
        name='127.0.0.1',
        url=dlaurl[:-1],
    )
    # sub
    assert_in_results(
        res,
        action='create_sibling_webdav.storage',
        status='ok',
        type='sibling',
        name='127.0.0.1-storage',
        path=sub.path,
        url=f'{url}/3f7',
    )
    assert_in_results(
        res,
        action='create_sibling_webdav',
        status='notneeded',
        type='sibling',
        name='127.0.0.1',
        path=sub.path,
    )
    # sub2
    assert_in_results(
        res,
        action='create_sibling_webdav.storage',
        status='notneeded',
        type='sibling',
        name='127.0.0.1-storage',
        path=sub2.path
    )
    assert_in_results(
        res,
        action='create_sibling_webdav',
        status='ok',
        type='sibling',
        path=sub2.path,
        name='127.0.0.1',
        url=f'{dlaurl}sub2',
    )
    # subsub
    assert_in_results(
        res,
        action='create_sibling_webdav.storage',
        status='notneeded',
        type='sibling',
        name='127.0.0.1-storage',
        path=subsub.path
    )
    assert_in_results(
        res,
        action='create_sibling_webdav',
        status='notneeded',
        type='sibling',
        name='127.0.0.1',
        path=subsub.path,
    )

    # should fail upfront with first discovered remote that already exist
    res = ds.create_sibling_webdav(
        url, mode='annex', existing='error', recursive=True,
        on_failure='ignore')
    assert_result_count(res, 8, status='error', type='sibling')
    # Note: 'message' is expected to be a tuple (and always present).
    assert all("is already configured" in r['message'][0] for r in res)
    assert all(r['action'].startswith('create_sibling_webdav') for r in res)

    srv_rt = Path(webdav_server.path)
    (srv_rt / '3f7').rmdir()
    (srv_rt / 'sub2' / 'subsub').rmdir()
    (srv_rt / 'sub2').rmdir()

    # existing=skip actually doesn't do anything (other than yielding notneeded)
    res = ds.create_sibling_webdav(url, mode='annex',
                                   existing='skip',
                                   recursive=True)
    assert_result_count(res, 8, status='notneeded')
    remote_content = list(srv_rt.glob('**'))
    assert len(remote_content) == 1  # nothing but root dir

    # reconfigure to move target one directory level:
    dlaurl += 'reconfigure'
    url += '/reconfigure'
    new_root = srv_rt / 'reconfigure'
    res = ds.create_sibling_webdav(url, mode='annex',
                                   existing='reconfigure',
                                   recursive=True)
    assert_result_count(res, 8, status='ok')
    assert all(r['action'].startswith('reconfigure_sibling_webdav')
               for r in res)
    remote_content = list(new_root.glob('**'))
    assert_in(new_root / '3f7', remote_content)
    assert_in(new_root / 'sub2', remote_content)
    assert_in(new_root / 'sub2' / 'subsub', remote_content)


def test_result_renderer(existing_dataset, credman,
                         webdav_credential, webdav_server):
    # need to amend the test credential, can only do after we know the URL
    # the test webdav webserver uses a realm label '/'
    credman.set(realm=f'{webdav_server.url}/', **webdav_credential)
    # consume stdout to make test self-contained
    out, err = run_main([
        'create-sibling-webdav',
        '-d', existing_dataset.path,
        webdav_server.url,
    ])
    # url is somehow reported
    assert_in('datalad-annex::?type=webdav', out)
    # and the two custom result renderings
    assert_in('create_sibling_webdav(ok)', out)
    assert_in('create_sibling_webdav.storage(ok)', out)



"""Handler for operations, such as "download", on ssh:// URLs"""

# allow for |-type UnionType declarations
from __future__ import annotations

import logging
import random
import sys
import time
from functools import partial
from math import floor
from pathlib import (
    Path,
    PurePosixPath,
)
from queue import (
    Full,
    Queue,
)
from typing import (
    Any,
    Dict,
    IO,
)
from urllib.parse import (
    urlparse,
    ParseResult,
)

from more_itertools import consume

from datalad_next.consts import COPY_BUFSIZE
from datalad_next.config import ConfigManager
from datalad_next.runners import CommandError
from datalad_next.shell import (
    FixedLengthResponseGeneratorPosix,
    ShellCommandExecutor,
    shell,
)

from .base import UrlOperations
from .exceptions import (
    UrlOperationsRemoteError,
    UrlOperationsResourceUnknown,
)

lgr = logging.getLogger('datalad.ext.next.ssh_url_operations')


__all__ = ['SshUrlOperations']


class SshUrlOperations(UrlOperations):
    """Handler for operations on ``ssh://`` URLs

    For downloading files, only servers that support execution of the commands
    'ls -dln', 'awk', and 'cat' are supported. This includes a wide
    range of operating systems, including devices that provide these commands
    via the 'busybox' software.

    .. note::
       Any instance of ``SshUrlOperations`` must be deleted before ending the
       program, otherwise python might not exit. The reason is, that
       ``SshUrlOperations`` retains and reuses SSH connections for subsequent
       command execution. Each connection has two threads associated with it.
       Those threads are only terminated when the connection is closed. The
       destructor of ``SshUrlOperations`` closes all connections and terminates
       all associated threads.
    """
    def __init__(self, *, cfg: ConfigManager | None = None):
        super().__init__(cfg=cfg)
        self.ssh_shells: dict[tuple[str, ...], tuple[ShellCommandExecutor, Any]] = dict()

    def __del__(self):
        for ssh_executor, context in self.ssh_shells.values():
            ssh_executor.close()
            context.__exit__(None, None, None)

    @staticmethod
    def _check_return_code(return_code: int | None, url: str, msg: str = ''):
        if return_code == 244:
            # this is the special code for a file-not-found
            raise UrlOperationsResourceUnknown(url, message=msg)
        elif return_code != 0:
            raise UrlOperationsRemoteError(
                url,
                message=f'ssh command returned {return_code}'
                + f': {msg}' if msg else ''
            )

    def ssh_shell_for(self,
                      url: str) -> ShellCommandExecutor:
        """Get a ShellCommandExecutor for the url (cached or newly created)"""
        open_args = ssh_url2openargs(url, self.cfg)[0]
        key = tuple(open_args)
        if key not in self.ssh_shells:
            context = shell(['ssh'] + open_args)
            try:
                ssh_executor = context.__enter__()
            except CommandError as e:
                context.__exit__(None, None, None)
                raise UrlOperationsRemoteError(url) from e
            self.ssh_shells[key] = (ssh_executor, context)
        return self.ssh_shells[key][0]

    def close_shell_for(self, url: str):
        """Close the ShellCommandExecutor for the url and remove it"""
        open_args = ssh_url2openargs(url, self.cfg)[0]
        key = tuple(open_args)
        if key in self.ssh_shells:
            ssh_executor, context = self.ssh_shells.pop(key)
            ssh_executor.close()
            context.__exit__(None, None, None)

    def stat(self,
             url: str,
             *,
             credential: str | None = None,
             timeout: float | None = None) -> Dict:
        """Gather information on a URL target, without downloading it

        See :meth:`datalad_next.url_operations.UrlOperations.stat`
        for parameter documentation and exception behavior.
        """
        # Check whether a readable file exists at the path. If not signal a
        # dedicated 244 return code. This allows the user to distinguish the
        # absence of a readable file from other errors, e.g. from an error in
        # awk. Only a missing file would yield 244. A ssh-connection problem
        # would lead to a 255 error (and a closed connection).
        stat_cmd = """
            ret() {{ return $1; }}
            test -r {fpath}
            if [ $? -eq 0 ]; then
                LC_ALL=C ls -dln -- {fpath} | awk '{{print $5; exit}}'
            else
                ret 244
            fi"""

        cmd = self.format_cmd(stat_cmd, url)
        ssh = self.ssh_shell_for(url)
        result = ssh(cmd)
        self._check_return_code(result.returncode, url, result.stderr.decode())
        return {'content-length': int(result.stdout)}

    def delete(self,
               url: str,
               *,
               credential: str | None = None,
               timeout: float | None = None) -> Dict:
        """Delete the target of a shh://-URL

        The target can be a file or a directory. `delete` will attempt to
        delete write protected targets (by setting write permissions). If
        the target is a directory, the complete directory and all its
        content will be deleted. `delete` will not modify the permissions
        of the parent of the target. That means, it will not delete a target
        in a write protected directory, but it will empty target, if target is
        a directory.

        See :meth:`datalad_next.url_operations.UrlOperations.delete`
        for parameter documentation and exception behavior.

        Raises
        ------
        UrlOperationsResourceUnknown
          For deletion targets found absent.
        """

        delete_cmd = """
            ret() {{ return $1; }}
            if [ -f {fpath} ]; then
                chmod u+w {fpath}
                rm -f {fpath}
            elif [ -d {fpath} ]; then
                chmod -R u+wx {fpath}
                rm -rf {fpath}
            else
                ret 244
            fi"""

        cmd = self.format_cmd(delete_cmd, url)
        ssh = self.ssh_shell_for(url)
        result = ssh(cmd)
        self._check_return_code(result.returncode, url, result.stderr.decode())
        return {}

    def download(self,
                 from_url: str,
                 to_path: Path | None,
                 *,
                 # unused, but theoretically could be used to
                 # obtain escalated/different privileges on a system
                 # to gain file access
                 credential: str | None = None,
                 hash: list[str] | None = None,
                 timeout: float | None = None) -> Dict:
        """Download a file by streaming it through an SSH connection.

        On the server-side, the file size is determined and sent. Afterwards
        the file content is sent via `cat` to the SSH client.

        See :meth:`datalad_next.url_operations.UrlOperations.download`
        for parameter documentation and exception behavior.
        """
        hasher = self._get_hasher(hash)
        progress_id = self._get_progress_id(from_url, str(to_path))

        # get the size of the file to download
        stat = self.stat(from_url, credential=credential, timeout=timeout)
        expected_size = stat['content-length']

        # get a shell command executor and a fixed length response generator
        ssh = self.ssh_shell_for(from_url)
        response_generator = FixedLengthResponseGeneratorPosix(
            ssh.stdout,
            expected_size
        )

        dst_fp = sys.stdout.buffer \
            if to_path is None \
            else open(to_path, 'wb')

        # Localize variable access to minimize overhead
        dst_fp_write = dst_fp.write

        # We already know that file exists, so we can just cat it.
        cmd = self.format_cmd('cat {fpath}', from_url)
        result_generator = ssh.start(
            cmd,
            response_generator=response_generator
        )
        # We do not use the `shell.operations.posix.download`-method here
        # because we need access to every individual chunk in order to calculate
        # the hash on the fly.
        for chunk in self._with_progress(
                result_generator,
                progress_id=progress_id,
                label='downloading',
                expected_size=expected_size,
                start_log_msg=('Download %s to %s', from_url, to_path),
                end_log_msg=('Finished download',),
                update_log_msg=('Downloaded chunk',)
        ):
            # write data
            dst_fp_write(chunk)
            # compute hash simultaneously
            hasher.update(chunk)

        if dst_fp and to_path is not None:
            dst_fp.close()

        self._check_return_code(
            result_generator.returncode,
            from_url,
            ''.join(result_generator.stderr_deque)
        )

        return {
            **stat,
            **hasher.get_hexdigest(),
        }

    def upload(self,
               from_path: Path | None,
               to_url: str,
               *,
               credential: str | None = None,
               hash: list[str] | None = None,
               timeout: float | None = None) -> Dict:
        """Upload a file by streaming it through an SSH connection.

        It, more or less, runs `ssh <host> 'cat > <path>'` or
        `ssh <host> 'head -c <file-size> > <path>'` on the remote side.

        See :meth:`datalad_next.url_operations.UrlOperations.upload`
        for parameter documentation and exception behavior.
        """

        if from_path is None:
            source_name = '<STDIN>'
            return self._perform_upload(
                src_fp=sys.stdin.buffer,
                source_name=source_name,
                to_url=to_url,
                hash_names=hash,
                expected_size=None,
                timeout=timeout,
            )
        else:
            # die right away, if we lack read permissions or there is no file
            with from_path.open("rb") as src_fp:
                return self._perform_upload(
                    src_fp=src_fp,
                    source_name=str(from_path),
                    to_url=to_url,
                    hash_names=hash,
                    expected_size=from_path.stat().st_size,
                    timeout=timeout,
                )

    def _perform_upload(self,
                        src_fp: IO,
                        source_name: str,
                        to_url: str,
                        hash_names: list[str] | None,
                        expected_size: int | None,
                        timeout: float | None) -> dict:

        hasher = self._get_hasher(hash_names)

        # we use a queue to implement timeouts.
        # we limit the queue to few items in order to `make queue.put()`
        # block relatively quickly, and thereby have the progress report
        # actually track the upload, i.e. the feeding of the stdin pipe
        # of the ssh-process, and not just the feeding of the
        # queue.
        # If we did not support timeouts, we could just use the following
        # as `input`-iterable for `iter_subproc`:
        #
        #   `iter(partial(src_fp.read, COPY_BUFSIZE), b'')
        #
        upload_queue: Queue = Queue(maxsize=2)

        if expected_size:
            read_cmd = f"head -c {expected_size}"
        else:
            read_cmd = "cat"

        cmd = self.format_cmd(
            # copy the file to its destination location with a randomized
            # name, and move it to its final location after upload. This
            # way, upload appears atomic, i.e. no half uploaded file will
            # be seen at the destination URL
            # leave special exit code when writing or moving fails, but not
            # the general SSH access
            "ret() {{ return $1; }}; ( mkdir -p '{fdir}' "
            f"&& {read_cmd} "
            "> '{fpath}.transfer-{nonce}' "
            "&& mv '{fpath}.transfer-{nonce}' '{fpath}' ) || ret 243",
            to_url,
        )

        progress_id = self._get_progress_id(source_name, to_url)

        ssh = self.ssh_shell_for(to_url)
        result_generator = ssh.start(
            cmd,
            stdin=self._with_progress(
                iter(upload_queue.get, None),
                progress_id=progress_id,
                label='uploading',
                expected_size=expected_size,
                start_log_msg=('Upload %s to %s', source_name, to_url),
                end_log_msg=('Finished upload',),
                update_log_msg=('Uploaded chunk',)
            )
        )

        try:
            upload_size = 0
            for chunk in iter(partial(src_fp.read, COPY_BUFSIZE), b''):

                # we are just putting stuff in the queue, and rely on
                # its maxsize to cause it to block the next call to
                # have the progress reports be anyhow valid, we also
                # rely on put-timeouts to implement timeout.
                upload_queue.put(chunk, timeout=timeout)

                # compute hash simultaneously
                hasher.update(chunk)
                upload_size += len(chunk)

            upload_queue.put(None, timeout=timeout)

        except Full:
            # we had a timeout while uploading
            raise TimeoutError(f'timeout while executing: {cmd}')

        if expected_size:
            consume(result_generator)
        else:
            # If the remote shell reads from stdin, its stdin has to be close
            # for the upload-command to terminate
            if expected_size is None:
                ssh.close()
            consume(result_generator)
            # stdin of the shell was closed, it cannot be used anymore.
            self.close_shell_for(to_url)

        self._check_return_code(
            result_generator.returncode,
            to_url,
            ''.join(result_generator.stderr_deque)
        )

        return {
            **hasher.get_hexdigest(),
            # return how much was copied. we could compare with
            # `expected_size` and error on mismatch, but not all
            # sources can provide that (e.g. stdin)
            'content-length': upload_size
        }

    def format_cmd(self,
                   cmd: str,
                   url: str) -> str:
        ssh_command_builder = _SshCommandBuilder(url, self.cfg)
        return ssh_command_builder.substitute(cmd)


class _SshCommandBuilder:
    def __init__(
            self,
            url: str,
            cfg: ConfigManager,
    ):
        self.ssh_args, self._parsed = ssh_url2openargs(url, cfg)
        self.ssh_args.extend(('-e', 'none'))
        # make sure the essential pieces exist
        assert self._parsed.path
        time_stamp = time.time()
        self.substitutions = dict(
            fdir=str(PurePosixPath(self._parsed.path).parent),
            fpath=self._parsed.path,
            nonce=(
                str(random.randint(1000000000, 9999999999))
                + '_'
                + str(time_stamp - floor(time_stamp))[2:0]
            )
        )

    def substitute(self, payload_cmd: str) -> str:
        return payload_cmd.format(**self.substitutions)


def ssh_url2openargs(
    url: str,
    cfg: ConfigManager,
) -> tuple[list[str], ParseResult]:
    """Helper to report ssh-open arguments from a URL and config

    Returns a tuple with the argument list and the parsed URL.
    """
    args: list[str] = list()
    parsed = urlparse(url)
    # make sure the essential pieces exist
    assert parsed.hostname
    for opt, arg in (('-p', parsed.port),
                     ('-l', parsed.username),
                     ('-i', cfg.get('datalad.ssh.identityfile'))):
        if arg:
            # f-string, because port is not str
            args.extend((opt, f'{arg}'))
    # we could also use .netloc here and skip -p/-l above
    args.append(parsed.hostname)
    return args, parsed



"""Utilities and types for collection iterators"""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
import os
from pathlib import (
    Path,
    PurePath,
)
import stat
from typing import (
    TYPE_CHECKING,
    Dict,
    Union,
    Any,
    IO,
    List,
)

from datalad_next.consts import COPY_BUFSIZE
from datalad_next.utils import MultiHash

if TYPE_CHECKING:
    from .annexworktree import AnnexWorktreeFileSystemItem
    from .directory import DirectoryItem
    from .gitworktree import GitWorktreeFileSystemItem
    from io import BufferedReader
    from os import PathLike
    from tarfile import ExFileObject
    from zipfile import ZipExtFile


# TODO Could be `StrEnum`, came with PY3.11
class FileSystemItemType(Enum):
    """Enumeration of file system path types

    The associated ``str`` values are chosen to be appropriate for
    downstream use (e.g, as type labels in DataLad result records).
    """
    file = 'file'
    directory = 'directory'
    symlink = 'symlink'
    hardlink = 'hardlink'
    specialfile = 'specialfile'


@dataclass
class NamedItem:
    name: Any


@dataclass
class TypedItem:
    type: Any


@dataclass
class PathBasedItem(NamedItem):
    """An item with a path as its ``name``

    A dedicated property supports the conversion of the
    native name representation into a ``PurePath`` instance.

    Any argument understood by the ``PurePath`` constructor can
    be used as ``name``, such as a a filename, a relative path, or an
    absolute path -- in string form, as path segment sequence, or
    a ``Path`` instance.

    It is recommended to use name/path values that are relative
    to the containing collection (directory, archive, repository, etc.).
    """
    def path(self) -> PurePath:
        """Returns the item name as a ``PurePath`` instance

        This default implementation assumes the ``name`` to be
        platform path conventions.
        """
        return PurePath(self.name)


@dataclass  # sadly PY3.10+ only (kw_only=True)
class FileSystemItem(PathBasedItem, TypedItem):
    type: FileSystemItemType
    size: int
    mtime: float | None = None
    mode: int | None = None
    uid: int | None = None
    gid: int | None = None
    link_target: str | PathLike[str] | None = None
    fp: IO | None = None

    def link_target_path(self) -> PurePath | None:
        """Returns the link_target as a ``PurePath`` instance"""
        return PurePath(self.link_target) if self.link_target is not None \
            else None

    @classmethod
    def from_path(
        cls,
        path: Path,
        *,
        link_target: bool = True,
    ) -> Union[
        DirectoryItem,
        AnnexWorktreeFileSystemItem,
        FileSystemItem,
        GitWorktreeFileSystemItem,
    ]:
        """Populate item properties from a single `stat` and `readlink` call

        The given ``path`` must exist. The ``link_target`` flag indicates
        whether to report the result of ``readlink`` for a symlink-type
        path.
        """
        cstat = path.lstat()
        cmode = cstat.st_mode
        if stat.S_ISLNK(cmode):
            ctype = FileSystemItemType.symlink
        elif stat.S_ISDIR(cmode):
            ctype = FileSystemItemType.directory
        else:
            # the rest is a file
            # there could be fifos and sockets, etc.
            # but we do not recognize them here
            ctype = FileSystemItemType.file
        item = cls(
            name=path,
            type=ctype,
            size=cstat.st_size,
            mode=cmode,
            mtime=cstat.st_mtime,
            uid=cstat.st_uid,
            gid=cstat.st_gid,
        )
        if link_target and ctype == FileSystemItemType.symlink:
            # could be p.readlink() from PY3.9+
            # but check performance difference
            item.link_target = os.readlink(path)
        return item


def compute_multihash_from_fp(
    fp: Union[BufferedReader, ExFileObject, ZipExtFile],
    hash: List[str],
    bufsize: int = COPY_BUFSIZE,
) -> Dict[str, str]:
    """Compute multiple hashes from a file-like
    """
    mhash = MultiHash(hash)
    while True:
        chunk = fp.read(bufsize)
        if not chunk:
            break
        mhash.update(chunk)
    return mhash.get_hexdigest()



from datalad.api import clone
from datalad_next.tests import skip_if_on_windows


# we cannot yet run on windows. see
# https://github.com/datalad/datalad-next/issues/654
def test_ria_ssh_roundtrip(
        sshserver, existing_dataset, no_result_rendering,
        tmp_path):
    ds = existing_dataset
    sshurl, sshlocalpath = sshserver
    testfile = ds.pathobj / 'testfile1.txt'
    testfile_content = 'uppytyup!'
    testfile.write_text(testfile_content)
    ds.save()
    # create store
    ds.create_sibling_ria(
        f'ria+{sshurl}',
        name='datastore',
        new_store_ok=True,
    )
    # push to store
    ds.push(to='datastore')
    # clone from store into a new location
    dsclone = clone(
        source=f'ria+{sshurl}#{ds.id}',
        path=tmp_path,
    )
    dsclone.get('.')
    assert ds.id == dsclone.id
    assert (ds.pathobj / 'testfile1.txt').read_text() \
        == (dsclone.pathobj / 'testfile1.txt').read_text() \
        == 'uppytyup!'



"""Handlers for operations on various URL types and protocols

Available handlers:

.. currentmodule:: datalad_next.url_operations
.. autosummary::
   :toctree: generated

   UrlOperations
   AnyUrlOperations
   FileUrlOperations
   HttpUrlOperations
   SshUrlOperations
   UrlOperationsRemoteError
   UrlOperationsResourceUnknown
   UrlOperationsInteractionError
   UrlOperationsAuthenticationError
   UrlOperationsAuthorizationError
"""

from .base import (
    # base class for 3rd-party extensions and implementations
    UrlOperations,
)

# operation support for different protocols
from .any import AnyUrlOperations
from .file import FileUrlOperations
from .http import HttpUrlOperations
from .ssh import SshUrlOperations

# primary exceptions types
from .exceptions import (
    UrlOperationsRemoteError,
    UrlOperationsResourceUnknown,
    UrlOperationsInteractionError,
    UrlOperationsAuthenticationError,
    UrlOperationsAuthorizationError,
)

# TODO REMOVE EVERYTHING BELOW FOR V2.0
import logging
from functools import partial
from more_itertools import side_effect
from pathlib import Path
from typing import (
    Any,
    Dict,
    Generator,
    Iterable,
)

import datalad
from datalad_next.config import ConfigManager
from datalad_next.utils import log_progress
from datalad_next.utils.multihash import (
    MultiHash,
    NoOpHash,
)

lgr = logging.getLogger('datalad.ext.next.url_operations')



"""Recognize DATALAD_TESTS_TMP_KEYRING_PATH to set alternative secret storage

Within `pytest` DataLad uses the plaintext keyring backend. This backend has no
built-in way to configure a custom file location for secret storage from the
outside. This patch looks for a DATALAD_TESTS_TMP_KEYRING_PATH environment
variable, and uses its value as a file path for the storage.

This makes it possible to (temporarily) switch storage. This feature is used
by the ``tmp_keyring`` pytest fixture. This patch is needed in addition to the
test fixture in order to apply such changes also to child processes, such as
special remotes and git remotes.
"""

from os import environ

if 'DATALAD_TESTS_TMP_KEYRING_PATH' in environ:
    import keyring
    kr = keyring.get_keyring()
    kr.file_path = environ['DATALAD_TESTS_TMP_KEYRING_PATH']



"""Uniform pre-execution parameter validation for commands

With this patch commands can now opt-in to receive fully validated parameters.
This can substantially simplify the implementation complexity of a command at
the expense of a more elaborate specification of the structural and semantic
properties of the parameters.

For details on implementing validation for individual commands see
:class:`datalad_next.commands.ValidatedInterface`.
"""

import logging
from typing import (
    Callable,
    Dict,
    Generator,
)

from datalad import cfg as dlcfg
from datalad.core.local.resulthooks import (
    get_jsonhooks_from_config,
    match_jsonhook2result,
    run_jsonhook,
)
from datalad.interface.results import known_result_xfms
from datalad.interface.utils import (
    anInterface,
    get_result_filter,
    keep_result,
    render_action_summary,
    xfm_result,
    _process_results,
)
from datalad_next.exceptions import IncompleteResultsError
from . import apply_patch
from datalad_next.constraints import DatasetParameter

# use same logger as -core
lgr = logging.getLogger('datalad.interface.utils')


# this is a replacement for datalad.interface.base.get_allargs_as_kwargs
# it reports which arguments were at their respective defaults
def get_allargs_as_kwargs(call, args, kwargs):
    """Generate a kwargs dict from a call signature and ``*args``, ``**kwargs``

    Basically resolving the argnames for all positional arguments, and
    resolving the defaults for all kwargs that are not given in a kwargs
    dict

    Returns
    -------
    (dict, set, set)
      The first return value is a mapping of argument names to their respective
      values.
      The second return value in the tuple is a set of argument names for
      which the effective value is identical to the default declared in the
      signature of the callable.
      The third value is a set with names of all mandatory arguments, whether
      or not they are included in the returned mapping.
    """
    from datalad_next.utils import getargspec
    argspec = getargspec(call, include_kwonlyargs=True)
    defaults = argspec.defaults
    nargs = len(argspec.args)
    defaults = defaults or []  # ensure it is a list and not None
    assert (nargs >= len(defaults))
    # map any args to their name
    argmap = list(zip(argspec.args[:len(args)], args))
    kwargs_ = dict(argmap)
    # map defaults of kwargs to their names (update below)
    default_map = dict(zip(argspec.args[-len(defaults):], defaults))
    for k, v in default_map.items():
        if k not in kwargs_:
            kwargs_[k] = v
    # update with provided kwarg args
    kwargs_.update(kwargs)
    # determine which arguments still have values identical to their declared
    # defaults
    at_default = set(
        k for k in kwargs_
        if k in default_map and default_map[k] == kwargs_[k]
    )
    # XXX we cannot assert the following, because our own highlevel
    # API commands support more kwargs than what is discoverable
    # from their signature...
    #assert (nargs == len(kwargs_))
    return (
        # argument name/value mapping
        kwargs_,
        # names of arguments that are at their default
        at_default,
        # names of mandatory arguments (set for uniformity)
        set(argspec.args),
    )


# This function interface is taken from
# datalad-core@209bc319db8f34cceae4fee86493bf41927676fd
def _execute_command_(
    *,
    interface: anInterface,
    cmd: Callable[..., Generator[Dict, None, None]],
    cmd_args: tuple,
    cmd_kwargs: Dict,
    exec_kwargs: Dict,
) -> Generator[Dict, None, None]:
    """Internal helper to drive a command execution generator-style

    Parameters
    ----------
    interface:
      Interface class of associated with the `cmd` callable
    cmd:
      A DataLad command implementation. Typically the `__call__()` of
      the given `interface`.
    cmd_args:
      Positional arguments for `cmd`.
    cmd_kwargs:
      Keyword arguments for `cmd`.
    exec_kwargs:
      Keyword argument affecting the result handling.
      See `datalad.interface.common_opts.eval_params`.
    """
    # for result filters and validation
    # we need to produce a dict with argname/argvalue pairs for all args
    # incl. defaults and args given as positionals
    allkwargs, at_default, required_args = get_allargs_as_kwargs(
        cmd,
        cmd_args,
        {**cmd_kwargs, **exec_kwargs},
    )

    # validate the complete parameterization
    param_validator = interface.get_parameter_validator() \
        if hasattr(interface, 'get_parameter_validator') else None
    if param_validator is None:
        lgr.debug(
            'Command parameter validation skipped. %s declares no validator',
            interface)
    else:
        lgr.debug('Command parameter validation for %s', interface)
        validator_kwargs = dict(
            at_default=at_default,
            required=required_args or None,
        )
        # make immediate vs exhaustive parameter validation
        # configurable
        raise_on_error = dlcfg.get(
            'datalad.runtime.parameter-violation', None)
        if raise_on_error:
            validator_kwargs['on_error'] = raise_on_error

        allkwargs = param_validator(
            allkwargs,
            **validator_kwargs
        )
        lgr.debug('Command parameter validation ended for %s', interface)

    # look for potential override of logging behavior
    result_log_level = dlcfg.get('datalad.log.result-level', 'debug')
    # resolve string labels for transformers too
    result_xfm = known_result_xfms.get(
        allkwargs['result_xfm'],
        # use verbatim, if not a known label
        allkwargs['result_xfm'])
    result_filter = get_result_filter(allkwargs['result_filter'])
    result_renderer = allkwargs['result_renderer']
    if result_renderer == 'tailored' and not hasattr(interface,
                                                     'custom_result_renderer'):
        # a tailored result renderer is requested, but the class
        # does not provide any, fall back to the generic one
        result_renderer = 'generic'
    if result_renderer == 'default':
        # standardize on the new name 'generic' to avoid more complex
        # checking below
        result_renderer = 'generic'

    # figure out which hooks are relevant for this command execution
    # query cfg for defaults
    # .is_installed and .config can be costly, so ensure we do
    # it only once. See https://github.com/datalad/datalad/issues/3575
    dataset_arg = allkwargs.get('dataset', None)
    ds = None
    if dataset_arg is not None:
        from datalad_next.datasets import Dataset
        if isinstance(dataset_arg, Dataset):
            ds = dataset_arg
        elif isinstance(dataset_arg, DatasetParameter):
            ds = dataset_arg.ds
        else:
            try:
                ds = Dataset(dataset_arg)
            except ValueError:
                pass
    # look for hooks
    hooks = get_jsonhooks_from_config(ds.config if ds else dlcfg)
    # end of hooks discovery

    # flag whether to raise an exception
    incomplete_results = []
    # track what actions were performed how many times
    action_summary = {}

    # if a custom summary is to be provided, collect the results
    # of the command execution
    results = []
    do_custom_result_summary = result_renderer in (
        'tailored', 'generic', 'default') and hasattr(
            interface,
            'custom_result_summary_renderer')
    pass_summary = do_custom_result_summary \
        and getattr(interface,
                    'custom_result_summary_renderer_pass_summary',
                    None)

    # process main results
    for r in _process_results(
            # execution, call with any arguments from the validated
            # set that are no result-handling related
            cmd(**{k: v for k, v in allkwargs.items()
                if k not in exec_kwargs}),
            interface,
            allkwargs['on_failure'],
            # bookkeeping
            action_summary,
            incomplete_results,
            # communication
            result_renderer,
            result_log_level,
            # let renderers get to see how a command was called
            allkwargs):
        for hook, spec in hooks.items():
            # run the hooks before we yield the result
            # this ensures that they are executed before
            # a potentially wrapper command gets to act
            # on them
            if match_jsonhook2result(hook, r, spec['match']):
                lgr.debug('Result %s matches hook %s', r, hook)
                # a hook is also a command that yields results
                # so yield them outside too
                # users need to pay attention to void infinite
                # loops, i.e. when a hook yields a result that
                # triggers that same hook again
                for hr in run_jsonhook(hook, spec, r, dataset_arg):
                    # apply same logic as for main results, otherwise
                    # any filters would only tackle the primary results
                    # and a mixture of return values could happen
                    if not keep_result(hr, result_filter, **allkwargs):
                        continue
                    hr = xfm_result(hr, result_xfm)
                    # rationale for conditional is a few lines down
                    if hr:
                        yield hr
        if not keep_result(r, result_filter, **allkwargs):
            continue
        r = xfm_result(r, result_xfm)
        # in case the result_xfm decided to not give us anything
        # exclude it from the results. There is no particular reason
        # to do so other than that it was established behavior when
        # this comment was written. This will not affect any real
        # result record
        if r:
            yield r

        # collect if summary is desired
        if do_custom_result_summary:
            results.append(r)

    # result summary before a potential exception
    # custom first
    if do_custom_result_summary:
        if pass_summary:
            summary_args = (results, action_summary)
        else:
            summary_args = (results,)
        interface.custom_result_summary_renderer(*summary_args)
    elif result_renderer in ('generic', 'default') \
            and action_summary \
            and sum(sum(s.values())
                    for s in action_summary.values()) > 1:
        # give a summary in generic mode, when there was more than one
        # action performed
        render_action_summary(action_summary)

    if incomplete_results:
        raise IncompleteResultsError(
            failed=incomplete_results,
            msg="Command did not complete successfully")


# apply patch
patch_msg = \
    'Apply datalad-next patch to interface.(utils|base).py:_execute_command_'
apply_patch('datalad.interface.base', None, '_execute_command_',
            _execute_command_, msg=patch_msg)



# ex: set sts=4 ts=4 sw=4 noet:
# -*- coding: utf-8 -*-
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""

"""
import pytest

from datalad_next.config import ConfigManager
from ..manager import (
    CredentialManager,
    _get_cred_cfg_var,
)
from datalad_next.tests import (
    assert_in,
    assert_raises,
    eq_,
)
from datalad_next.utils import chpwd


def test_credmanager(tmp_keyring, datalad_cfg, datalad_interactive_ui):
    ui = datalad_interactive_ui
    credman = CredentialManager(datalad_cfg)
    # doesn't work with thing air
    assert_raises(ValueError, credman.get)
    eq_(credman.get('donotexiststest'), None)
    # we get reports as soon as there is a secret available
    # this makes it possible to discover credential fragments, if only to
    # expose them for clean-up
    eq_(credman.get(crazy='empty'), None)
    eq_(credman.get(crazy='empty', secret='bogus'),
        {'crazy': 'empty', 'secret': 'bogus'})
    # does not fiddle with a secret that is readily provided
    eq_(credman.get('dummy', secret='mike', _type_hint='token'),
        dict(type='token', secret='mike'))

    # remove a secret that has not yet been set
    eq_(credman.remove('elusive', type_hint='user_password'), False)

    # but the secret was written to the keystore
    with pytest.raises(ValueError):
        credman.set('mycred', forbidden_propname='some')

    # no instructions what to do, no legacy entry, nothing was changed
    # but the secret was written to the keystore
    eq_(credman.set('mycred', secret='some'), dict(secret='some'))
    # redo but with timestep
    setprops = credman.set('lastusedcred', _lastused=True, secret='some')
    assert_in('last-used', setprops)
    # now re-set, based on the retrieved info, but update the timestamp
    setprops_new = credman.set('lastusedcred', _lastused=True,
                               **credman.get('lastusedcred'))
    # must have updated 'last-used'
    assert setprops['last-used'] != setprops_new['last-used']
    # first property store attempt
    eq_(credman.set('changed', secret='some', prop='val'),
        dict(secret='some', prop='val'))
    # second, no changing the secret, but changing the prop, albeit with
    # the same value, change report should be empty
    eq_(credman.set('changed', prop='val'), dict())
    # change secret, with value pulled from config
    try:
        datalad_cfg.set('datalad.credential.changed.secret', 'envsec',
                        scope='override')
        eq_(credman.set('changed', secret=None), dict(secret='envsec'))
    finally:
        datalad_cfg.unset('datalad.credential.changed.secret',
                          scope='override')

    # remove non-existing property, secret not report, because unchanged
    eq_(credman.set('mycred', dummy=None), dict(dummy=None))
    assert _get_cred_cfg_var("mycred", "dummy") not in datalad_cfg

    # set property
    eq_(credman.set('mycred', dummy='good', this='that'),
        dict(dummy='good', this='that'))
    # ensure set
    eq_(credman.get('mycred'), dict(dummy='good', this='that', secret='some'))
    # remove individual property
    eq_(credman.set('mycred', dummy=None), dict(dummy=None))
    # remove individual property that is not actually present
    eq_(credman.set('mycred', imaginary=None), dict(imaginary=None))
    # ensure removal
    eq_(credman.get('mycred'), dict(this='that', secret='some'))

    # test full query and constrained query
    q = list(credman.query_())
    # 3 defined here, plus any number of legacy credentials
    assert len(q) > 3
    # now query for one of the creds created above
    q = list(credman.query_(prop='val'))
    eq_(len(q), 1)
    eq_(q[0][0], 'changed')
    eq_(q[0][1]['prop'], 'val')
    # and now a query with no match
    q = list(credman.query_(prop='val', funky='town'))
    eq_(len(q), 0)

    # remove complete credential
    credman.remove('mycred')
    eq_(credman.get('mycred'), None)

    # test prompting for a secret when none is given
    ui.staged_responses.append('mysecret')
    res = credman.set('mycred', other='prop')
    assert res == {'other': 'prop', 'secret': 'mysecret'}

    # test prompting for a name when None is given
    ui.staged_responses.append('mycustomname')
    res = credman.set(None, secret='dummy', other='prop')
    assert res == {'name': 'mycustomname', 'other': 'prop', 'secret': 'dummy'}

    # test name prompt loop in case of a name collision
    ui.staged_responses.extend(['mycustomname', 'mycustomname2'])
    res = credman.set(None, secret='dummy2', other='prop2')
    assert res == {'name': 'mycustomname2', 'other': 'prop2',
                   'secret': 'dummy2'}

    # test skipping at prompt, smoke test _context arg
    ui.staged_responses.append('skip')
    res = credman.set(
        None, _context='for me', secret='dummy', other='prop')
    assert res is None

    # accept suggested name
    ui.staged_responses.append('')
    res = credman.set(
        None, _suggested_name='auto1', secret='dummy', other='prop')
    assert res == {'name': 'auto1', 'other': 'prop', 'secret': 'dummy'}

    # a suggestion conflicting with an existing credential is like
    # not making a suggestion at all
    ui.staged_responses.extend(('', 'auto2'))
    res = credman.set(
        None, _suggested_name='auto1', secret='dummy', other='prop')
    assert res == {'name': 'auto2', 'other': 'prop', 'secret': 'dummy'}


def test_credmanager_set_noninteractive(
        tmp_keyring, datalad_cfg, datalad_noninteractive_ui):
    credman = CredentialManager(datalad_cfg)
    # if no name is provided and none _can_ be entered -> raise
    with pytest.raises(ValueError):
        credman.set(None, secret='dummy', other='prop')


def test_credman_local(existing_dataset):
    ds = existing_dataset
    credman = CredentialManager(ds.config)

    # deposit a credential into the dataset's config, and die trying to
    # remove it
    ds.config.set('datalad.credential.stupid.secret', 'really', scope='branch')
    assert_raises(RuntimeError, credman.remove, 'stupid')

    # but it manages for the local scope
    ds.config.set('datalad.credential.notstupid.secret', 'really', scope='local')
    credman.remove('notstupid')


def test_query(tmp_keyring, datalad_cfg):
    credman = CredentialManager(datalad_cfg)
    # set a bunch of credentials with a common realm AND timestamp
    for i in range(3):
        credman.set(
            f'cred.{i}',
            _lastused=True,
            secret=f'diff{i}',
            realm='http://ex.com/login',
        )
    # now a credential with the common realm, but without a timestamp
    credman.set(
        'cred.no.time',
        _lastused=False,
        secret='notime',
        realm='http://ex.com/login',
    )
    # and the most recent one (with timestamp) is an unrelated one
    credman.set('unrelated', _lastused=True, secret='unrelated')

    # smoke test for an unsorted report
    assert len(credman.query()) > 1
    # now we want all credentials that match the realm, sorted by
    # last-used timestamp -- most recent first
    slist = credman.query(realm='http://ex.com/login', _sortby='last-used')
    eq_(['cred.2', 'cred.1', 'cred.0', 'cred.no.time'],
        [i[0] for i in slist])
    # same now, but least recent first, importantly no timestamp stays last
    slist = credman.query(realm='http://ex.com/login', _sortby='last-used',
                          _reverse=False)
    eq_(['cred.0', 'cred.1', 'cred.2', 'cred.no.time'],
        [i[0] for i in slist])


def test_credman_get(datalad_cfg, datalad_interactive_ui):
    ui = datalad_interactive_ui
    # we are not making any writes, any config must work
    credman = CredentialManager(datalad_cfg)
    # must be prompting for missing properties
    ui.staged_responses.append('myuser')
    res = credman.get(
        None, _type_hint='user_password', _prompt='myprompt',
        secret='dummy')
    assert 'myuser' == res['user']
    # same for the secret
    ui.staged_responses.append('mysecret')
    res = credman.get(
        None, _type_hint='user_password', _prompt='myprompt',
        user='dummy')
    assert 'mysecret' == res['secret']


def test_credman_get_guess_type():
    # define token-only-no-type credential in config override
    credman = CredentialManager(
        ConfigManager(overrides={
            'datalad.credential.mike.token': 'some',
        })
    )
    # we get it reported fine, token property converted to the
    # 'secret' and a proper 'type' assigned
    assert credman.get('mike') == {
        'secret': 'some',
        'type': 'token',
    }


def test_credman_obtain(tmp_keyring, datalad_cfg, datalad_interactive_ui):
    ui = datalad_interactive_ui
    credman = CredentialManager(datalad_cfg)
    # senseless, but valid call
    # could not possibly report a credential without any info
    with pytest.raises(ValueError):
        credman.obtain()
    # a type_hint is not enough, if no prompt is provided
    with pytest.raises(ValueError):
        credman.obtain(type_hint='token')
    # also a prompt alone is not enough
    with pytest.raises(ValueError):
        credman.obtain(prompt='myprompt')
    # minimal condition prompt and type-hint for manual entry
    ui.staged_responses.append('mytoken')
    res = credman.obtain(type_hint='token', prompt='myprompt')
    assert res == (None,
                   {'type': 'token', 'secret': 'mytoken', '_edited': True})

    # no place a credential we could discover
    cred1_props = dict(secret='sec1', type='token', realm='myrealm')
    credman.set('cred1', _lastused=True, **cred1_props)

    # one matching property is all that is needed
    res = credman.obtain(query_props={'realm': 'myrealm'})
    assert res == ('cred1', credman.get('cred1'))

    # will report the last-used one
    credman.set('cred2', _lastused=True, **cred1_props)
    res = credman.obtain(query_props={'realm': 'myrealm'})
    assert res == ('cred2', credman.get('cred2'))
    credman.set('cred1', _lastused=True, **cred1_props)
    res = credman.obtain(query_props={'realm': 'myrealm'})
    assert res == ('cred1', credman.get('cred1'))

    # built-in test for additional property expectations
    with pytest.raises(ValueError):
        credman.obtain(query_props={'realm': 'myrealm'},
                       expected_props=['funky'])

    res = credman.obtain(query_props={'realm': 'myrealm'})
    # if we are looking for a realm, we get it back even if a credential
    # had to be entered
    ui.staged_responses.append('mynewtoken')
    res = credman.obtain(
        type_hint='token', prompt='myprompt',
        query_props={'realm': 'mytotallynewrealm'})
    assert res == (None,
                   {'type': 'token', 'secret': 'mynewtoken', '_edited': True,
                    'realm': 'mytotallynewrealm'})


legacy_provider_cfg = """\
[provider:credmanuniquetestcredentialsetup]
url_re = http://example\\.com/
authentication_type = http_basic_auth
credential = credmanuniquetestcredentialsetup

[credential:credmanuniquetestcredentialsetup]
type = user_password
"""


def test_legacy_credentials(tmp_keyring, existing_dataset):
    # - the legacy code will only ever pick up a dataset credential, when
    #   PWD is inside a dataset
    # - we want all tests to bypass the actual system keyring
    #   'datalad.downloaders.credentials.keyring_' is what the UserPassword
    #   credential will use to store the credential
    # - 'datalad.support.keyring_' is what credman uses
    # - we need to make them one and the same thing, and the tmp_keyring
    #   fixture does this by replacing the keyring storage for the runtime
    #   of the test
    with chpwd(existing_dataset.path):
        check_legacy_credentials(tmp_keyring, existing_dataset)


def check_legacy_credentials(tmp_keyring, existing_dataset):
    # we will use a dataset to host a legacy provider config
    ds = existing_dataset
    provider_path = ds.pathobj / '.datalad' / 'providers' / 'mylegacycred.cfg'
    provider_path.parent.mkdir(parents=True, exist_ok=True)
    provider_path.write_text(legacy_provider_cfg)

    # shortcut
    cname = 'credmanuniquetestcredentialsetup'

    credman = CredentialManager(ds.config)
    # check that we get legacy reports in a query. this is needed to be
    # able to even know that they exist
    res = dict(credman.query())
    assert cname in res
    cred = res[cname]
    # we always get the type reported
    assert cred['type'] == 'user_password'
    # we can know that it is a legacy credential
    assert cred['_from_backend'] == 'legacy'
    # but getting an unset legacy credential will unambiguously say
    # "there is none"
    assert credman.get(cname) is None

    # we want all tests to bypass the actual system keyring
    # 'datalad.downloaders.credentials.keyring_' is what the UserPassword
    # credential will use to store the credential
    from datalad.downloaders.credentials import UserPassword
    lc = UserPassword(cname, dataset=ds)
    lc.set(user='mike', password='pass123')
    # now we should be able to get it from credman too
    # and just by name -- no need to provide a type hint
    cred = credman.get(cname)
    assert cred['user'] == 'mike'
    # reporting of the secret is always under the 'secret' key
    assert cred['secret'] == 'pass123'
    assert cred['type'] == 'user_password'
    assert cred['_from_backend'] == 'legacy'

    # check migration on set
    try:
        # setting a credential, will migrate info into the non-legacy
        # backend. however, it will not move information _out of_
        # the legacy backend, in order to keep old code working
        # with the old info
        # confirm starting point: legacy code keeps user in secret store
        assert tmp_keyring.get_password(f'datalad-{cname}', 'user') == 'mike'
        assert ds.config.get(f'datalad.credential.{cname}.user') is None
        credman.set(cname, **cred)
        # it remains there
        assert tmp_keyring.get_password(f'datalad-{cname}', 'user') == 'mike'
        # but is also migrated
        assert ds.config.get(f'datalad.credential.{cname}.user') == 'mike'
        # NOTE: This setup is not without problems. Users will update
        # a credential and will leave an outdated (half) as garbage.
        # however, I did not come up with a better approach that gradually
        # brings users over.
        credman.set(cname, user='newmike', secret='othersecret')
        assert tmp_keyring.get_password(f'datalad-{cname}', 'user') == 'mike'
        # finally check that the update is reported now
        cred = credman.get(cname)
        assert cred['user'] == 'newmike'
        assert cred['secret'] == 'othersecret'
        assert cred['type'] == 'user_password'
        # no legacy info makes it out, hence no marker
        assert cred.get('_from_backend') != 'legacy'
    finally:
        credman.remove(cname)



"""Enable ``configuration()`` to query ``global`` scope without a dataset

"""

__docformat__ = 'restructuredtext'


import logging

from datalad import cfg as dlcfg
from datalad.distribution.dataset import require_dataset
from datalad_next.commands import (
    build_doc,
    datasetmethod,
    eval_results,
    get_status_dict,
)
from datalad.interface.common_cfg import definitions as cfg_defs
from datalad.local import configuration as conf_mod
from datalad.local.configuration import (
    config_actions,
    _dump,
    _get,
    _set,
    _unset,
)
from datalad_next.exceptions import NoDatasetFound
from datalad_next.utils import ensure_list
from datalad_next.datasets import (
    Dataset,
)

lgr = logging.getLogger('datalad.local.configuration')


@build_doc
class Configuration(conf_mod.Configuration):
    """"""
    @staticmethod
    @datasetmethod(name='configuration')
    @eval_results
    def __call__(
            action='dump',
            spec=None,
            *,
            scope=None,
            dataset=None,
            recursive=False,
            recursion_limit=None):

        # check conditions
        # - global and recursion makes no sense

        if action == 'dump':
            if scope:
                raise ValueError(
                    'Scope selection is not supported for dumping')

        # normalize variable specifications
        specs = []
        for s in ensure_list(spec):
            if isinstance(s, tuple):
                specs.append((str(s[0]), str(s[1])))
            elif '=' not in s:
                specs.append((str(s),))
            else:
                specs.append(tuple(s.split('=', 1)))

        if action == 'set':
            missing_values = [s[0] for s in specs if len(s) < 2]
            if missing_values:
                raise ValueError(
                    'Values must be provided for all configuration '
                    'settings. Missing: {}'.format(missing_values))
            invalid_names = [s[0] for s in specs if '.' not in s[0]]
            if invalid_names:
                raise ValueError(
                    'Name must contain a section (i.e. "section.name"). '
                    'Invalid: {}'.format(invalid_names))

        ds = None
        if scope != 'global' or recursive:
            try:
                ds = require_dataset(
                    dataset,
                    check_installed=True,
                    purpose='configure')
            except NoDatasetFound:
                if action not in ('dump', 'get') or dataset:
                    raise

        res_kwargs = dict(
            action='configuration',
            logger=lgr,
        )
        if ds:
            res_kwargs['refds'] = ds.path
        yield from configuration(action, scope, specs, res_kwargs, ds)

        if not recursive:
            return

        for subds in ds.subdatasets(
                state='present',
                recursive=True,
                recursion_limit=recursion_limit,
                on_failure='ignore',
                return_type='generator',
                result_renderer='disabled'):
            yield from configuration(
                action, scope, specs, res_kwargs, Dataset(subds['path']))


def configuration(action, scope, specs, res_kwargs, ds=None):
    # go with the more specific dataset configmanager, if we are
    # operating on a dataset
    cfg = dlcfg if ds is None else ds.config

    if action not in config_actions:
        raise ValueError("Unsupported action '{}'".format(action))

    if action == 'dump':
        if not specs:
            # dumping is querying for all known keys
            specs = [
                (n,) for n in sorted(
                    set(cfg_defs.keys()).union(cfg.keys()))
            ]
        scope = None

    for spec in specs:
        if '.' not in spec[0]:
            yield get_status_dict(
                ds=ds,
                status='error',
                message=(
                    "Configuration key without a section: '%s'",
                    spec[0],
                ),
                **res_kwargs)
            continue
        # TODO without get-all there is little sense in having add
        #if action == 'add':
        #    res = _add(cfg, scope, spec)
        if action == 'get':
            res = _get(cfg, scope, spec[0])
            # `None` is a value that cannot be set in the config.
            # if it is returned, it indicates that no value was set
            # we need to communicate that back, because a None value
            # cannot be reported as such by the CLI
            # (only via, e.g. JSON, encoding).
            # It makes sense to communicate that getting this specific
            # configuration item is "impossible" (because it is not set).
            # if a caller wants to tollerate this scenario, they can
            # set on_failure='ignore'
            if res.get('value') is None:
                res['status'] = 'impossible'
                res['message'] = (
                    'key %r not set in configuration%s',
                    res['name'],
                    f" scope '{scope}'" if scope else '',
                )
        elif action == 'dump':
            res = _dump(cfg, spec[0])
        # TODO this should be there, if we want to be comprehensive
        # however, we turned this off by default in the config manager
        # because we hardly use it, and the handling in ConfigManager
        # is not really well done.
        #elif action == 'get-all':
        #    res = _get_all(cfg, scope, spec)
        elif action == 'set':
            res = _set(cfg, scope, *spec)
        elif action == 'unset':
            res = _unset(cfg, scope, spec[0])

        if ds:
            res['path'] = ds.path

        if 'status' not in res:
            res['status'] = 'ok'

        yield dict(res_kwargs, **res)

    if action in ('add', 'set', 'unset'):
        # we perform a single reload, rather than one for each modification
        # TODO: can we detect a call from cmdline? We could skip the reload.
        cfg.reload(force=True)


conf_mod.Configuration.__call__ = Configuration.__call__
conf_mod.Configuration._params_['scope']._doc = """\
    scope for getting or setting
    configuration. If no scope is declared for a query, all
    configuration sources (including overrides via environment
    variables) are considered according to the normal
    rules of precedence. A 'get' action can be constrained to
    scope 'branch', otherwise 'global' is used when not operating
    on a dataset, or 'local' (including 'global', when operating
    on a dataset.
    For action 'dump', a scope selection is ignored and all available
    scopes are considered."""
conf_mod.Configuration.__call__.__doc__ = None
conf_mod.Configuration = build_doc(conf_mod.Configuration)



import requests

from ..requests_auth import DataladAuth


def test_token_credential_uses_bearer_with_basic_only_challenge():
    response = requests.Response()
    response.status_code = 401
    response.url = 'https://example.com/file'
    response.headers['www-authenticate'] = 'Basic realm=""'

    auth = DataladAuth.__new__(DataladAuth)

    def get_credential(url, auth_schemes):
        assert url == response.url
        assert auth_schemes == {'basic': {'realm': ''}}
        return None, 'mytoken', {
            'type': 'token',
            'secret': 'sekrit',
        }

    rerequest = {}

    def authenticated_rerequest(response_, request_auth, **kwargs):
        rerequest['response'] = response_
        prep = requests.Request('GET', response_.url).prepare()
        request_auth(prep)
        return prep

    auth._get_credential = get_credential
    auth._authenticated_rerequest = authenticated_rerequest

    renewed_request = auth.handle_401(response)

    assert rerequest['response'] is response
    assert renewed_request.headers['Authorization'] == 'Bearer sekrit'



from datalad.conftest import setup_package

# fixture setup
from datalad_next.tests.fixtures import (
    # no test can leave global config modifications behind
    check_gitconfig_global,
    # no test can leave secrets behind
    check_plaintext_keyring,
    # function-scope credential manager
    credman,
    # function-scope config manager
    datalad_cfg,
    # function-scope UI wrapper that can provide staged responses
    datalad_interactive_ui,
    # function-scope UI wrapper that can will raise when asked for responses
    datalad_noninteractive_ui,
    # function-scope temporary keyring
    tmp_keyring,
    # function-scope, Dataset instance
    dataset,
    # function-scope, Dataset instance with underlying repository
    existing_dataset,
    # function-scope, Dataset instance with underlying Git-only repository
    existing_noannex_dataset,
    # session-scope, Dataset instance with various modifications,
    # to-be-treated read-only
    modified_dataset,
    # session-scope, standard http credential (full dict)
    http_credential,
    # function-scope, auth-less HTTP server
    http_server,
    # function-scope, HTTP server with required authentication
    http_server_with_basicauth,
    # function-scope relay httpbin_service, unless undesired and skips instead
    httpbin,
    # session-scope HTTPBIN instance startup and URLs
    httpbin_service,
    # function-scope, disabled datalad command result rendering for all
    # command calls
    no_result_rendering,
    # session-scope redirection of log messages
    reduce_logging,
    # session-scope determine setup of an SSH server to use for testing
    sshserver_setup,
    # function-scope SSH server base url and local path
    sshserver,
    # session-scope, standard webdav credential (full dict)
    webdav_credential,
    # function-scope, serve a local temp-path via WebDAV
    webdav_server,
)
from datalad_next.iter_collections.tests.test_itertar import (
    # session-scope, downloads a tarball with a set of standard
    # file/dir/link types
    sample_tar_xz,
)
from datalad_next.iter_collections.tests.test_iterzip import (
    # session-scope, create a sample zip file
    sample_zip,
)



"""Execution of subprocesses

This module provides all relevant components for subprocess execution.  The
main work horse is :func:`~datalad_next.runners.iter_subproc`, a context
manager that enables interaction with a subprocess in the form of an iterable
for input/output processing. Execution errors are communicated with the
:class:`~datalad_next.runners.CommandError` exception. In addition, a few
convenience functions are provided to execute Git commands (including
git-annex).

.. currentmodule:: datalad_next.runners
.. autosummary::
   :toctree: generated

   iter_subproc
   call_git
   call_git_lines
   call_git_oneline
   call_git_success
   iter_git_subproc
   CommandError


Low-level tooling from datalad-core
-----------------------------------

.. deprecated:: 1.4
   The functionality described here has been deprecated, and the associated
   imports from datalad-core are scheduled for removal with version 2.0.
   Use the implementations listed above instead.

Few process execution/management utilities are provided, for
generic command execution, and for execution command in the context
of a Git repository.

.. autosummary::
   :toctree: generated

   GitRunner
   Runner

Additional information on the design of the subprocess execution tooling
is available from https://docs.datalad.org/design/threaded_runner.html

A standard exception type is used to communicate any process termination
with a non-zero exit code

.. autosummary::
   :toctree: generated

   CommandError

Command output can be processed via "protocol" implementations that are
inspired by ``asyncio.SubprocessProtocol``.

.. autosummary::
   :toctree: generated

   KillOutput
   NoCapture
   StdOutCapture
   StdErrCapture
   StdOutErrCapture
"""

from .iter_subproc import (
    iter_subproc,
)
from .git import (
    call_git,
    call_git_lines,
    call_git_oneline,
    call_git_success,
    iter_git_subproc,
)

# runners
# TODO REMOVE FOR V2.0
from datalad.runner import (
    GitRunner,
    Runner,
)
# TODO REMOVE FOR V2.0
from datalad.runner.nonasyncrunner import ThreadedRunner
# protocols
# TODO REMOVE FOR V2.0
from datalad.runner import (
    KillOutput,
    NoCapture,
    Protocol,
    StdOutCapture,
    StdErrCapture,
    StdOutErrCapture,
)
# TODO REMOVE FOR V2.0
from datalad.runner.protocol import GeneratorMixIn
# TODO REMOVE FOR V2.0
from .protocols import (
    NoCaptureGeneratorProtocol,
    StdOutCaptureGeneratorProtocol,
)
# exceptions
# The following import supports legacy code that uses `CommandError` from this
# module. If you are writing new code, please use `CommandError` from
# `datalad.support.exceptions`. We intend to remove this import in the future.
from datalad_next.exceptions import CommandError

# utilities
# TODO REMOVE FOR V2.0
from datalad.runner.nonasyncrunner import (
    STDOUT_FILENO,
    STDERR_FILENO,
)
# TODO REMOVE FOR V2.0
from datalad.runner.utils import (
    LineSplitter,
)
# TODO REMOVE FOR V2.0
from subprocess import (
    DEVNULL,
)



from itertools import chain
import pytest

from datalad_next.datasets import Dataset
from datalad_next.runners import (
    call_git_success,
)

from ..gitstatus import (
    GitDiffStatus,
    GitContainerModificationType,
    iter_gitstatus,
)


def test_status_homogeneity(modified_dataset):
    """Test things that should always be true, no matter the precise
    parameterization

    A main purpose of this test is also to exercise all (main) code paths.
    """
    ds = modified_dataset
    for kwargs in (
        # default
        dict(path=ds.pathobj),
        dict(path=ds.pathobj, recursive='no'),
        dict(path=ds.pathobj, recursive='repository'),
        dict(path=ds.pathobj, recursive='submodules'),
        # same as above, but with the submodules in the root
        dict(path=ds.pathobj / 'dir_sm', recursive='no'),
        dict(path=ds.pathobj / 'dir_sm', recursive='repository'),
        dict(path=ds.pathobj / 'dir_sm', recursive='submodules'),
        # no submodule state
        dict(path=ds.pathobj, eval_submodule_state='no', recursive='no'),
        dict(path=ds.pathobj, eval_submodule_state='no', recursive='repository'),
        dict(path=ds.pathobj, eval_submodule_state='no', recursive='submodules'),
        # just the commit
        dict(path=ds.pathobj, eval_submodule_state='commit', recursive='no'),
        dict(path=ds.pathobj, eval_submodule_state='commit', recursive='repository'),
        dict(path=ds.pathobj, eval_submodule_state='commit', recursive='submodules'),
        # without untracked
        dict(path=ds.pathobj, untracked=None, recursive='no'),
        dict(path=ds.pathobj, untracked=None, recursive='repository'),
        dict(path=ds.pathobj, untracked=None, recursive='submodules'),
        # special untracked modes
        dict(path=ds.pathobj, untracked='whole-dir', recursive='no'),
        dict(path=ds.pathobj, untracked='whole-dir', recursive='repository'),
        dict(path=ds.pathobj, untracked='whole-dir', recursive='submodules'),
        dict(path=ds.pathobj, untracked='no-empty-dir', recursive='no'),
        dict(path=ds.pathobj, untracked='no-empty-dir', recursive='repository'),
        dict(path=ds.pathobj, untracked='no-empty-dir', recursive='submodules'),
        # call in the mountpoint of a dropped submodule
        dict(path=ds.pathobj / 'dir_sm' / 'droppedsm_c'),
    ):
        st = {item.name: item for item in iter_gitstatus(**kwargs)}
        # we get no report on anything clean (implicitly also tests
        # whether all item names are plain strings
        assert all(not i.name.endswith('_c') for i in st.values())

        # anything untracked is labeled as such
        assert all(
            i.status == GitDiffStatus.other
            # we would not see a submodule modification qualifier when instructed
            # not to evaluate a submodule
            or kwargs.get('eval_submodule_state') in ('no', 'commit')
            or GitContainerModificationType.untracked_content in i.modification_types
            for i in st.values()
            if 'u' in i.path.name.split('_')[1]
        )

        # anything modified is labeled as such
        assert all(
            # either directly
            i.status == GitDiffStatus.modification
            # or as an addition with a modification on top
            or (i.status == GitDiffStatus.addition
                and GitContainerModificationType.modified_content
                    in i.modification_types)
            for i in st.values()
            if 'm' in i.path.name.split('_')[1]
        )

        # anything deleted is labeled as such
        assert all(
            i.status == GitDiffStatus.deletion
            for i in st.values()
            if 'd' in i.path.name.split('_')[1]
        )


def test_status_invalid_params(existing_dataset):
    ds = existing_dataset
    with pytest.raises(ValueError):
        list(iter_gitstatus(ds.pathobj, recursive='fromspace'))


test_cases_repository_recursion = [
    {'name': 'file_a', 'status': GitDiffStatus.addition},
    {'name': 'dir_m/file_a', 'status': GitDiffStatus.addition},
    {'name': 'file_u', 'status': GitDiffStatus.other},
    {'name': 'dir_u/file_u', 'status': GitDiffStatus.other},
    {'name': 'dir_m/file_u', 'status': GitDiffStatus.other},
    {'name': 'dir_m/dir_u/file_u', 'status': GitDiffStatus.other},
    {'name': 'file_d', 'status': GitDiffStatus.deletion},
    {'name': 'dir_d/file_d', 'status': GitDiffStatus.deletion},
    {'name': 'dir_m/file_d', 'status': GitDiffStatus.deletion},
    {'name': 'file_m', 'status': GitDiffStatus.modification},
    {'name': 'dir_m/file_m', 'status': GitDiffStatus.modification},
    {'name': 'dir_sm/sm_d', 'status': GitDiffStatus.deletion},
    {'name': 'dir_sm/sm_n', 'status': GitDiffStatus.modification,
     'qual': (GitContainerModificationType.new_commits,)},
    {'name': 'dir_sm/sm_m', 'status': GitDiffStatus.modification,
     'qual': (GitContainerModificationType.modified_content,)},
    {'name': 'dir_sm/sm_nm', 'status': GitDiffStatus.modification,
     'qual': (GitContainerModificationType.modified_content,
              GitContainerModificationType.new_commits)},
    {'name': 'dir_sm/sm_nmu', 'status': GitDiffStatus.modification,
     'qual': (GitContainerModificationType.modified_content,
              GitContainerModificationType.untracked_content,
              GitContainerModificationType.new_commits)},
    {'name': 'dir_sm/sm_u', 'status': GitDiffStatus.modification,
     'qual': (GitContainerModificationType.untracked_content,)},
    {'name': 'dir_sm/sm_mu', 'status': GitDiffStatus.modification,
     'qual': (GitContainerModificationType.modified_content,
              GitContainerModificationType.untracked_content)},
]

test_cases_submodule_recursion = [
    {'name': 'dir_sm/sm_m/file_a', 'status': GitDiffStatus.addition},
    {'name': 'dir_sm/sm_nm/file_a', 'status': GitDiffStatus.addition},
    {'name': 'dir_sm/sm_mu/file_a', 'status': GitDiffStatus.addition},
    {'name': 'dir_sm/sm_nmu/file_a', 'status': GitDiffStatus.addition},
    {'name': 'dir_sm/sm_m/file_m', 'status': GitDiffStatus.modification},
    {'name': 'dir_sm/sm_mu/file_m', 'status': GitDiffStatus.modification},
    {'name': 'dir_sm/sm_nmu/file_m', 'status': GitDiffStatus.modification},
    {'name': 'dir_sm/sm_u/file_u', 'status': GitDiffStatus.other},
    {'name': 'dir_sm/sm_mu/file_u', 'status': GitDiffStatus.other},
    {'name': 'dir_sm/sm_nmu/file_u', 'status': GitDiffStatus.other},
    {'name': 'dir_sm/sm_u/dir_u/file_u', 'status': GitDiffStatus.other},
    {'name': 'dir_sm/sm_mu/dir_u/file_u', 'status': GitDiffStatus.other},
    {'name': 'dir_sm/sm_nmu/dir_u/file_u', 'status': GitDiffStatus.other},
]


def _assert_testcases(st, tc):
    for c in tc:
        assert st[c['name']].status == c['status']
        mod_types = st[c['name']].modification_types
        if 'qual' in c:
            assert set(mod_types) == set(c['qual']), \
                f'Mismatch for {c=!r}'
        else:
            assert mod_types is None


def test_status_vs_git(modified_dataset):
    """Implements a comparison against how git-status behaved when
    the test was written  (see fixture docstring)
    """
    st = {
        item.name: item
        for item in iter_gitstatus(
            path=modified_dataset.pathobj, recursive='repository',
            eval_submodule_state='full', untracked='all',
        )
    }
    _assert_testcases(st, test_cases_repository_recursion)


def test_status_norec(modified_dataset):
    st = {
        item.name: item
        for item in iter_gitstatus(
            path=modified_dataset.pathobj, recursive='no',
            eval_submodule_state='full', untracked='all',
        )
    }
    test_cases = [
        {'name': 'file_a', 'status': GitDiffStatus.addition},
        {'name': 'dir_d', 'status': GitDiffStatus.deletion},
        {'name': 'dir_m', 'status': GitDiffStatus.modification,
         'qual': (GitContainerModificationType.modified_content,
                  GitContainerModificationType.untracked_content)},
        {'name': 'dir_sm', 'status': GitDiffStatus.modification,
         'qual': (GitContainerModificationType.modified_content,
                  GitContainerModificationType.untracked_content)},
        {'name': 'file_d', 'status': GitDiffStatus.deletion},
        {'name': 'file_m', 'status': GitDiffStatus.modification},
        {'name': 'dir_u', 'status': GitDiffStatus.other},
        {'name': 'file_u', 'status': GitDiffStatus.other},
    ]
    _assert_testcases(st, test_cases)


def test_status_smrec(modified_dataset):
    st = {
        item.name: item
        for item in iter_gitstatus(
            path=modified_dataset.pathobj, recursive='submodules',
            eval_submodule_state='full', untracked='all',
        )
    }
    # in this mode we expect ALL results of a 'repository' mode recursion,
    # including the submodule-type items, plus additional ones from within
    # the submodules
    _assert_testcases(st, chain(test_cases_repository_recursion,
                                test_cases_submodule_recursion))


def test_status_monorec(modified_dataset):
    st = {
        item.name: item
        for item in iter_gitstatus(
            path=modified_dataset.pathobj, recursive='monolithic',
            eval_submodule_state='full', untracked='all',
        )
    }
    # in this mode we expect ALL results of a 'repository' mode recursion,
    # including the submodule-type items, plus additional ones from within
    # the submodules
    _assert_testcases(
        st,
        # repository and recursive test cases
        [c for c in chain(test_cases_repository_recursion,
                          test_cases_submodule_recursion)
         # minus any submodule that have no new commits
         # (this only thing that is not attributable to individual
         # content changes)
         if not c['name'].split('/')[-1] in (
             'sm_m', 'sm_mu', 'sm_u',
         )])


def test_status_gitinit(tmp_path):
    # initialize a fresh git repo, but make no commits
    assert call_git_success(['init'], cwd=tmp_path)
    for recmode in ('no', 'repository', 'submodules'):
        assert [] == list(iter_gitstatus(tmp_path, recursive=recmode))
    # untracked reporting must be working normal
    (tmp_path / 'untracked').touch()
    for recmode in ('no', 'repository', 'submodules'):
        res = list(iter_gitstatus(tmp_path, recursive=recmode))
        assert len(res) == 1
        assert res[0].name == 'untracked'
        assert res[0].status == GitDiffStatus.other


def test_status_nohead_staged(tmp_path):
    # initialize a fresh git repo, but make no commits
    assert call_git_success(['init'], cwd=tmp_path)
    # stage a file
    (tmp_path / 'probe').write_text('tostage')
    assert call_git_success(['add', 'probe'], cwd=tmp_path)
    _assert_testcases(
        {i.name: i for i in iter_gitstatus(tmp_path)},
        [{'name': 'probe', 'status': GitDiffStatus.addition}],
    )


def test_status_submodule_added(existing_dataset, no_result_rendering):
    p = existing_dataset.pathobj
    subm_name = 'sub'
    Dataset(p / subm_name).create()
    assert call_git_success(
        ['submodule', '--quiet', 'add', f'./{subm_name}', subm_name],
        cwd=p)
    # check that we get the status on a submodule that was just added with
    # no additional changes
    res = list(iter_gitstatus(p, recursive='monolithic'))
    assert any(i.name == subm_name and i.gitsha and i.prev_gitsha is None
               for i in res)



"""Enhance ``run()`` placeholder substitutions to honor configuration defaults

Previously, ``run()`` would not recognize configuration defaults for
placeholder substitution. This means that any placeholders globally declared in
``datalad.interface.common_cfg``, or via ``register_config()`` in DataLad
extensions would not be effective.

This patch makes run's ``format_command()`` helper include such defaults
explicitly, and thereby enable the global declaration of substitution defaults.

Moreoever a ``{python}`` placeholder is now defined via this mechanism, and
points to the value of ``sys.executable`` by default. This particular
placeholder was found to be valuable for improving the portability of
run-recording across (specific) Python versions, or across different (virtual)
environments. See https://github.com/datalad/datalad-container/issues/224 for
an example use case.

https://github.com/datalad/datalad/pull/7509
"""

from itertools import filterfalse
import sys

from datalad.core.local.run import (
    GlobbedPaths,
    SequenceFormatter,
    normalize_command,
    quote_cmdlinearg,
)
from datalad.interface.common_cfg import definitions as cfg_defs
from datalad.support.constraints import EnsureStr
from datalad.support.extensions import register_config

from . import apply_patch


# This function is taken from datalad-core@a96c51c0b2794b2a2b4432ec7bd51f260cb91a37
# datalad/core/local/run.py
# The change has been proposed in https://github.com/datalad/datalad/pull/7509
def format_command(dset, command, **kwds):
    """Plug in placeholders in `command`.

    Parameters
    ----------
    dset : Dataset
    command : str or list

    `kwds` is passed to the `format` call. `inputs` and `outputs` are converted
    to GlobbedPaths if necessary.

    Returns
    -------
    formatted command (str)
    """
    command = normalize_command(command)
    sfmt = SequenceFormatter()
    cprefix = 'datalad.run.substitutions.'

    def not_subst(x):
        return not x.startswith(cprefix)

    for k in set(filterfalse(not_subst, cfg_defs.keys())).union(
            filterfalse(not_subst, dset.config.keys())):
        v = dset.config.get(
            k,
            # pull a default from the config definitions
            # if we have no value, but a key
            cfg_defs.get(k, {}).get('default', None))
        sub_key = k.replace(cprefix, "")
        if sub_key not in kwds:
            kwds[sub_key] = v

    for name in ["inputs", "outputs"]:
        io_val = kwds.pop(name, None)
        if not isinstance(io_val, GlobbedPaths):
            io_val = GlobbedPaths(io_val, pwd=kwds.get("pwd"))
        kwds[name] = list(map(quote_cmdlinearg, io_val.expand(dot=False)))
    return sfmt.format(command, **kwds)


apply_patch(
    'datalad.core.local.run', None, 'format_command', format_command)
register_config(
    'datalad.run.substitutions.python',
    'Substitution for {python} placeholder',
    description='Path to a Python interpreter executable',
    type=EnsureStr(),
    default=sys.executable,
    dialog='question',
)



import io
import stat

import pytest
import sys

from datalad_next.tests import skip_if_on_windows

from ..file import (
    FileUrlOperations,
    UrlOperationsRemoteError,
    UrlOperationsResourceUnknown,
)


def test_file_url_download(tmp_path):
    test_path = tmp_path / 'myfile'
    test_url = test_path.as_uri()
    ops = FileUrlOperations()
    # no target file (yet), precise exception
    with pytest.raises(UrlOperationsResourceUnknown):
        ops.stat(test_url)
    # now put something at the target location
    test_path.write_text('surprise!')
    # and now it works
    props = ops.stat(test_url)
    # we get the correct file size reported
    assert props['content-length'] == test_path.stat().st_size

    # and download
    download_path = tmp_path / 'download'
    props = ops.download(test_url, download_path, hash=['sha256'])
    assert props['sha256'] == '71de4622cf536ed4aa9b65fc3701f4fc5a198ace2fa0bda234fd71924267f696'
    assert props['content-length'] == 9 == test_path.stat().st_size

    # remove source and try again
    test_path.unlink()
    with pytest.raises(UrlOperationsResourceUnknown):
        ops.download(test_url, download_path)


def test_file_url_upload(tmp_path, monkeypatch):
    payload = 'payload'
    payload_file = tmp_path / 'payload'
    test_upload_path = tmp_path / 'myfile'
    test_upload_url = test_upload_path.as_uri()
    ops = FileUrlOperations()
    # missing source file
    # standard exception, makes no sense to go custom thinks mih
    with pytest.raises(FileNotFoundError):
        ops.upload(payload_file, test_upload_url)
    # no empty targets lying around
    assert not test_upload_path.exists()

    # now again
    payload_file.write_text(payload)
    props = ops.upload(payload_file, test_upload_url, hash=['md5'])
    assert test_upload_path.read_text() == 'payload'
    assert props['content-length'] == len(payload)
    assert props['md5'] == '321c3cf486ed509164edec1e1981fec8'

    # upload from STDIN
    from_stdin_url = (tmp_path / 'missingdir' / 'from_stdin').as_uri()
    with monkeypatch.context() as m:
        m.setattr(sys, 'stdin',
                  io.TextIOWrapper(io.BytesIO(
                      bytes(payload, encoding='utf-8'))))
        props = ops.upload(None, from_stdin_url, hash=['md5'])
        assert props['md5'] == '321c3cf486ed509164edec1e1981fec8'
        assert props['content-length'] == len(payload)

    # TODO test missing write permissions


def test_file_url_delete(tmp_path):
    payload = 'payload'
    test_dir_path = tmp_path / 'subdir'
    test_path = test_dir_path / 'myfile'

    test_dir_path.mkdir()
    test_url = test_path.as_uri()
    test_dir_url = test_dir_path.as_uri()
    ops = FileUrlOperations()

    # missing file
    with pytest.raises(UrlOperationsResourceUnknown):
        ops.delete(test_url)

    # place file and write protect
    test_path.write_text(payload)
    assert test_path.read_text() == payload
    test_path.chmod(stat.S_IRUSR)
    # file deletion works
    ops.delete(test_url)
    assert not test_path.exists()

    # place file again protect directory and file
    test_path.write_text(payload)
    assert test_path.read_text() == payload
    test_dir_path.chmod(stat.S_IXUSR | stat.S_IRUSR)
    test_path.chmod(stat.S_IRUSR)
    # non-empty directory deletion works
    ops.delete(test_dir_url)
    assert not test_dir_path.exists()

    # create empty, write-protected dir
    assert not test_dir_path.exists()
    test_dir_path.mkdir()
    test_dir_path.chmod(stat.S_IXUSR | stat.S_IRUSR)
    ops.delete(test_dir_url)
    assert not test_dir_path.exists()


@skip_if_on_windows
def test_file_url_upload_errors(tmp_path):
    source = tmp_path / 'source'
    source.write_text('Some content\n')
    fops = FileUrlOperations()
    with pytest.raises(UrlOperationsRemoteError) as e:
        fops.upload(source, 'file:///tmp//')



"""TAR archive operation handler"""

# allow for |-type UnionType declarations
from __future__ import annotations

import logging
import tarfile
from contextlib import contextmanager
from pathlib import (
    Path,
    PurePosixPath,
)
from typing import (
    Generator,
    IO,
)

from datalad_next.config import ConfigManager
# TODO we might just want to do it in reverse:
# move the code of `iter_tar` in here and have it call
# `TarArchiveOperations(path).__iter__()` instead.
# However, the flexibility to have `iter_tar()` behave
# differently depending on parameters (fp=True/False)
# is nice, and `__iter__()` only has `self`, such that
# any customization would need to be infused in the whole
# class. Potentially cumbersome.
from datalad_next.iter_collections import (
    TarfileItem,
    iter_tar,
)

from .base import ArchiveOperations

lgr = logging.getLogger('datalad.ext.next.archive_operations.tarfile')


class TarArchiveOperations(ArchiveOperations):
    """Handler for a TAR archive on a local file system

    Any methods that take an archive item/member name as an argument
    accept a POSIX path string, or any `PurePath` instance.
    """
    def __init__(self, location: Path, *, cfg: ConfigManager | None = None):
        """
        Parameters
        ----------
        location: Path
          TAR archive location
        cfg: ConfigManager, optional
          A config manager instance that is consulted for any supported
          configuration items
        """
        # TODO expose `mode` other kwargs of `tarfile.TarFile`
        super().__init__(location, cfg=cfg)

        # Consider supporting file-like for `location`,
        # see tarfile.open(fileobj=)
        self._tarfile_path = location
        self._tarfile = None

    @property
    def tarfile(self) -> tarfile.TarFile:
        """Returns `TarFile` instance, after creating it on-demand

        The instance is cached, and needs to be released by calling
        ``.close()`` if called outside a context manager.
        """
        if self._tarfile is None:
            self._tarfile = tarfile.open(self._tarfile_path, 'r')
        return self._tarfile

    def close(self) -> None:
        """Closes any opened TAR file handler"""
        if self._tarfile:
            self._tarfile.close()
            self._tarfile = None

    @contextmanager
    def open(self, item: str | PurePosixPath) -> Generator[IO | None]:
        """Get a file-like for a TAR archive item

        The file-like object allows to read from the archive-item specified
        by `item`.

        Parameters
        ----------
        item: str | PurePath
          The identifier must be a POSIX path string, or a `PurePath` instance.

        Returns
        -------
        IO | None
          A file-like object to read bytes from the item, if the item is a
          regular file, else `None`. (This is returned by the context manager
          that is created via the decorator `@contextmanager`.)

        Raises
        ------
        KeyError
          If no item with the name `item` can be found in the tar-archive
        """
        with self.tarfile.extractfile(_anyid2membername(item)) as fp:
            yield fp

    def __contains__(self, item: str | PurePosixPath) -> bool:
        try:
            self.tarfile.getmember(_anyid2membername(item))
            return True
        except KeyError:
            return False

    def __iter__(self) -> Generator[TarfileItem, None, None]:
        # if fp=True is needed, either `iter_tar()` can be used
        # directly, or `TarArchiveOperations.open`
        yield from iter_tar(self._tarfile_path, fp=False)


def _anyid2membername(item_id: str | PurePosixPath) -> str:
    if isinstance(item_id, PurePosixPath):
        return item_id.as_posix()
    else:
        return item_id




import pytest

from .. import utils  # for patching environ

from ..utils import (
    get_gitconfig_items_from_env,
    set_gitconfig_items_in_env,
)


def test_get_gitconfig_items_from_env(monkeypatch):
    with monkeypatch.context() as m:
        # without the COUNT the rest does not matter and we always
        # get an empty dict
        m.delenv('GIT_CONFIG_COUNT', raising=False)
        assert get_gitconfig_items_from_env() == {}

    with monkeypatch.context() as m:
        # setting zero items, also makes everything else irrelevant
        m.setenv('GIT_CONFIG_COUNT', '0')
        assert get_gitconfig_items_from_env() == {}

    with monkeypatch.context() as m:
        # predictable error for botched count
        m.setenv('GIT_CONFIG_COUNT', 'rubbish')
        with pytest.raises(ValueError) as e:
            get_gitconfig_items_from_env()
        assert 'bogus count in GIT_CONFIG_COUNT' in str(e)

    # bunch of std error conditions
    for env, excstr in (
            ({'GIT_CONFIG_COUNT': 1,
              'GIT_CONFIG_KEY_0': 'section.name'},
             'missing config value'),
            ({'GIT_CONFIG_COUNT': 1,
              'GIT_CONFIG_VALUE_0': 'value'},
             'missing config key'),
            ({'GIT_CONFIG_COUNT': 1,
              'GIT_CONFIG_KEY_0': '',
              'GIT_CONFIG_VALUE_0': 'value'},
             'empty config key'),
            ({'GIT_CONFIG_COUNT': 1,
              'GIT_CONFIG_KEY_0': 'nosection',
              'GIT_CONFIG_VALUE_0': 'value'},
             'does not contain a section'),
    ):
        with monkeypatch.context() as m:
            m.setattr(utils, 'environ', env)
            with pytest.raises(ValueError) as e:
                get_gitconfig_items_from_env()
            assert excstr in str(e)

    # proper functioning
    for env, target in (
            ({'GIT_CONFIG_COUNT': 1,
              'GIT_CONFIG_KEY_0': 'section.name',
              'GIT_CONFIG_VALUE_0': 'value'},
             {'section.name': 'value'}),
            ({'GIT_CONFIG_COUNT': 2,
              'GIT_CONFIG_KEY_0': 'section.name1',
              'GIT_CONFIG_VALUE_0': 'value1',
              'GIT_CONFIG_KEY_1': 'section.name2',
              'GIT_CONFIG_VALUE_1': 'value2'},
             {'section.name1': 'value1', 'section.name2': 'value2'}),
            # double-specification appends
            # ❯ GIT_CONFIG_COUNT=2 \
            #   GIT_CONFIG_KEY_0=section.name \
            #   GIT_CONFIG_VALUE_0=val1 \
            # GIT_CONFIG_KEY_1=section.name \
            #   GIT_CONFIG_VALUE_1=val2 \
            #   git config --list --show-origin | grep 'command line:'
            # command line:   section.name=val1
            # command line:   section.name=val2
            ({'GIT_CONFIG_COUNT': 3,
              'GIT_CONFIG_KEY_0': 'section.name',
              'GIT_CONFIG_VALUE_0': 'value0',
              'GIT_CONFIG_KEY_1': 'section.name',
              'GIT_CONFIG_VALUE_1': 'value1',
              'GIT_CONFIG_KEY_2': 'section.name',
              'GIT_CONFIG_VALUE_2': 'value2'},
             {'section.name': ('value0', 'value1', 'value2')}),
    ):
        with monkeypatch.context() as m:
            m.setattr(utils, 'environ', env)
            assert get_gitconfig_items_from_env() == target


def test_set_gitconfig_items_in_env(monkeypatch):
    for start, items, target in (
            # giving nothing preserves statusquo
            ({}, {}, {}),
            ({'DUMMY': 'value'}, {}, {'DUMMY': 'value'}),
            # fixable specification is cleaned up
            ({'GIT_CONFIG_COUNT': '526'}, {}, {}),
            # but it has limits
            ({'GIT_CONFIG_COUNT': 'nochance'}, {},
             {'GIT_CONFIG_COUNT': 'nochance'}),
            # and there is no exhaustive search
            ({'GIT_CONFIG_KEY_3': 'dummy'}, {}, {'GIT_CONFIG_KEY_3': 'dummy'}),
            # virgin territory
            ({}, {'section.name': 'value'},
             {'GIT_CONFIG_COUNT': '1',
              'GIT_CONFIG_KEY_0': 'section.name',
              'GIT_CONFIG_VALUE_0': 'value'}),
            # "set" means "replace, not amend
            ({'GIT_CONFIG_COUNT': '1',
              'GIT_CONFIG_KEY_0': 'section.name',
              'GIT_CONFIG_VALUE_0': 'value'},
             {'altsection.name2': 'value2'},
             {'GIT_CONFIG_COUNT': '1',
              'GIT_CONFIG_KEY_0': 'altsection.name2',
              'GIT_CONFIG_VALUE_0': 'value2'}),
            # full cleanupage
            ({'GIT_CONFIG_COUNT': '2',
              'GIT_CONFIG_KEY_0': 'section.name',
              'GIT_CONFIG_VALUE_0': 'value',
              'GIT_CONFIG_KEY_1': 'altsection.name2',
              'GIT_CONFIG_VALUE_1': 'value2'},
             {}, {}),
            # multi-value support, order preserved
            ({}, {'section.name': ('c', 'a', 'b')},
             {'GIT_CONFIG_COUNT': '3',
              'GIT_CONFIG_KEY_0': 'section.name',
              'GIT_CONFIG_VALUE_0': 'c',
              'GIT_CONFIG_KEY_1': 'section.name',
              'GIT_CONFIG_VALUE_1': 'a',
              'GIT_CONFIG_KEY_2': 'section.name',
              'GIT_CONFIG_VALUE_2': 'b'}),
    ):
        with monkeypatch.context() as m:
            env = dict(start)
            m.setattr(utils, 'environ', env)
            set_gitconfig_items_in_env(items)
            assert env == target


def test_get_set_gitconfig_env_roundtrip(monkeypatch):
    items = {'section.name': ('c', 'a', 'b'),
             'space section.na me.so me': 'v al'}
    with monkeypatch.context() as m:
        env = {}
        m.setattr(utils, 'environ', env)
        # feed in copy to ensure validity of the test
        set_gitconfig_items_in_env(dict(items))
        assert get_gitconfig_items_from_env() == items
